# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import logging

import numpy as np
import pytest
import scipy.sparse as sparse

import pyttb as ttb
from pyttb.sptensor import tt_from_sparse_matrix, tt_to_sparse_matrix


@pytest.fixture()
def sample_sptensor():
    subs = np.array([[1, 1, 1], [1, 1, 3], [2, 2, 2], [3, 3, 3]])
    vals = np.array([[0.5], [1.5], [2.5], [3.5]])
    shape = (4, 4, 4)
    data = {"subs": subs, "vals": vals, "shape": shape}
    sptensorInstance = ttb.sptensor().from_data(subs, vals, shape)
    return data, sptensorInstance


@pytest.mark.indevelopment
def test_sptensor_initialization_empty():
    empty = np.array([], ndmin=2, dtype=int)

    # No args
    sptensorInstance = ttb.sptensor()
    assert (sptensorInstance.subs == empty).all()
    assert (sptensorInstance.vals == empty).all()
    assert (sptensorInstance.shape == empty).all()


@pytest.mark.indevelopment
def test_sptensor_initialization_from_data(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    assert (sptensorInstance.subs == data["subs"]).all()
    assert (sptensorInstance.vals == data["vals"]).all()
    assert sptensorInstance.shape == data["shape"]


@pytest.mark.indevelopment
def test_sptensor_initialization_from_tensor_type(sample_sptensor):
    # Copy constructor
    (data, sptensorInstance) = sample_sptensor
    sptensorCopy = ttb.sptensor.from_tensor_type(sptensorInstance)
    assert (sptensorCopy.subs == data["subs"]).all()
    assert (sptensorCopy.vals == data["vals"]).all()
    assert sptensorCopy.shape == data["shape"]

    # Convert Tensor
    inputData = np.array([[1, 2, 3], [4, 5, 6]])
    tensorInstance = ttb.tensor.from_data(inputData)
    sptensorFromTensor = ttb.sptensor.from_tensor_type(tensorInstance)
    logging.debug(f"inputData = {inputData}")
    logging.debug(f"tensorInstance = {tensorInstance}")
    logging.debug(f"sptensorFromTensor = {sptensorFromTensor}")
    assert (
        sptensorFromTensor.subs
        == ttb.tt_ind2sub(inputData.shape, np.arange(0, inputData.size))
    ).all()
    assert (
        sptensorFromTensor.vals == inputData.reshape((inputData.size, 1), order="F")
    ).all()
    assert sptensorFromTensor.shape == inputData.shape

    # From coo sparse matrix
    inputData = sparse.random(11, 4, 0.2)
    sptensorFromCOOMatrix = ttb.sptensor.from_tensor_type(sparse.coo_matrix(inputData))
    assert (sptensorFromCOOMatrix.spmatrix() != sparse.coo_matrix(inputData)).nnz == 0

    # Negative Tests
    with pytest.raises(AssertionError):
        invalid_tensor_type = []
        ttb.sptensor.from_tensor_type(invalid_tensor_type)


@pytest.mark.indevelopment
def test_sptensor_initialization_from_function():
    # Random Tensor Success
    def function_handle(*args):
        return np.array([[0.5], [1.5], [2.5], [3.5], [4.5], [5.5]])

    np.random.seed(123)
    shape = (4, 4, 4)
    nz = 6
    sptensorInstance = ttb.sptensor.from_function(function_handle, shape, nz)
    assert (sptensorInstance.vals == function_handle()).all()
    assert sptensorInstance.shape == shape
    assert len(sptensorInstance.subs) == nz

    # NZ as a propotion in [0,1)
    nz = 0.09375
    sptensorInstance = ttb.sptensor.from_function(function_handle, shape, nz)
    assert (sptensorInstance.vals == function_handle()).all()
    assert sptensorInstance.shape == shape
    assert len(sptensorInstance.subs) == int(nz * np.prod(shape))

    # Random Tensor exception for negative non-zeros
    nz = -1
    with pytest.raises(AssertionError) as excinfo:
        a = ttb.sptensor.from_function(function_handle, shape, nz)
    assert (
        "Requested number of non-zeros must be positive and less than the total size"
        in str(excinfo)
    )

    # Random Tensor exception for negative non-zeros
    nz = np.prod(shape) + 1
    with pytest.raises(AssertionError) as excinfo:
        a = ttb.sptensor.from_function(function_handle, shape, nz)
    assert (
        "Requested number of non-zeros must be positive and less than the total size"
        in str(excinfo)
    )


@pytest.mark.indevelopment
def test_sptensor_initialization_from_aggregator(sample_sptensor):
    subs = np.array([[1, 1, 1], [1, 1, 3], [2, 2, 2], [3, 3, 3], [1, 1, 1], [1, 1, 1]])
    vals = np.array([[0.5], [1.5], [2.5], [3.5], [4.5], [5.5]])
    shape = (4, 4, 4)
    a = ttb.sptensor.from_aggregator(subs, vals, shape)
    assert (a.subs == np.array([[1, 1, 1], [1, 1, 3], [2, 2, 2], [3, 3, 3]])).all()
    assert (a.vals == np.array([[10.5], [1.5], [2.5], [3.5]])).all()
    assert a.shape == shape

    a = ttb.sptensor.from_aggregator(subs, vals)
    assert (a.subs == np.array([[1, 1, 1], [1, 1, 3], [2, 2, 2], [3, 3, 3]])).all()
    assert (a.vals == np.array([[10.5], [1.5], [2.5], [3.5]])).all()
    assert a.shape == shape

    a = ttb.sptensor.from_aggregator(np.array([]), vals, shape)
    assert a.isequal(ttb.sptensor.from_data(np.array([]), np.array([]), shape))

    with pytest.raises(AssertionError) as excinfo:
        ttb.sptensor.from_aggregator(subs, np.concatenate((vals, np.array([[1.0]]))))
    assert "Number of subscripts and values must be equal" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        ttb.sptensor.from_aggregator(
            np.concatenate((subs, np.ones((6, 1))), axis=1), vals, shape
        )
    assert "More subscripts than specified by shape" in str(excinfo)

    badSubs = subs.copy()
    badSubs[0, 0] = 11
    with pytest.raises(AssertionError) as excinfo:
        ttb.sptensor.from_aggregator(badSubs, vals, shape)
    assert "Subscript exceeds sptensor shape" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor_and_scalar(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    b = sptensorInstance.logical_and(0)
    assert (b.subs == np.array([])).all()
    assert (b.vals == np.array([])).all()
    assert b.shape == data["shape"]

    b = sptensorInstance.logical_and(0.5)
    assert (b.subs == data["subs"]).all()
    assert (b.vals == np.array([[True], [False], [False], [False]])).all()
    assert b.shape == data["shape"]


@pytest.mark.indevelopment
def test_sptensor_and_sptensor(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    b = sptensorInstance.logical_and(sptensorInstance)

    assert (b.subs == data["subs"]).all()
    assert (b.vals == np.array([[True], [True], [True], [True]])).all()
    assert b.shape == data["shape"]

    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.logical_and(
            ttb.sptensor.from_data(data["subs"], data["vals"], (5, 5, 5))
        )
    assert "Must be tensors of the same shape" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.logical_and(np.ones(data["shape"]))
    assert "The arguments must be two sptensors or an sptensor and a scalar." in str(
        excinfo
    )


@pytest.mark.indevelopment
def test_sptensor_and_tensor(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    b = sptensorInstance.logical_and(ttb.tensor.from_tensor_type(sptensorInstance))
    assert (b.subs == data["subs"]).all()
    assert (b.vals == np.ones(data["vals"].shape)).all()


@pytest.mark.indevelopment
def test_sptensor_full(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    densetensor = sptensorInstance.full()
    denseData = np.zeros(sptensorInstance.shape)
    actualIdx = tuple(data["subs"].transpose())
    denseData[actualIdx] = data["vals"].transpose()[0]

    assert (densetensor.data == denseData).all()
    assert densetensor.shape == data["shape"]

    # Empty, no shape tensor conversion
    emptySptensor = ttb.sptensor()
    emptyTensor = ttb.tensor()
    assert emptyTensor.isequal(emptySptensor.full())

    # Empty, no non-zeros tensor conversion
    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), data["shape"])
    assert (emptySptensor.full().data == np.zeros(data["shape"])).all()


@pytest.mark.indevelopment
def test_sptensor_subdims(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    assert (sptensorInstance.subdims([[1], [1], [1, 3]]) == np.array([0, 1])).all()
    assert (
        sptensorInstance.subdims((1, 1, slice(None, None, None))) == np.array([0, 1])
    ).all()

    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.subdims([[1], [1, 3]])
    assert "Number of subdimensions must equal number of dimensions" in str(excinfo)

    with pytest.raises(ValueError):
        sptensorInstance.subdims(("bad", "region", "types"))


@pytest.mark.indevelopment
def test_sptensor_ndims(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    assert sptensorInstance.ndims == 3


@pytest.mark.indevelopment
def test_sptensor_extract(sample_sptensor, capsys):
    (data, sptensorInstance) = sample_sptensor

    # Out of range subs case
    # Too large
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.extract(np.array([[4, 4, 4], [1, 1, 1]]))
    assert "Invalid subscripts" in str(excinfo)
    capsys.readouterr()
    # Negative #TODO, would we like to support reverse indexing which is pythonic
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.extract(np.array([[-1, -1, -1], [1, 1, 1]]))
    assert "Invalid subscripts" in str(excinfo)
    capsys.readouterr()

    # List of subs case
    assert (
        sptensorInstance.extract(np.array([[1, 1, 1], [1, 1, 3]])) == [[0.5], [1.5]]
    ).all()

    # Single sub case
    # TODO if you pass a single sub should you get a list of vals with one entry or just a single val
    assert (sptensorInstance.extract(np.array([1, 1, 1])) == [[0.5]]).all()


@pytest.mark.indevelopment
def test_sptensor__getitem__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    ## Case 1
    # Empty value slice
    assert sptensorInstance[0, 0, 0] == 0
    # Full value slice
    assert sptensorInstance[1, 1, 1] == 0.5
    # Empty subtensor
    emptyResult = sptensorInstance[0:1, 0:1, 0:1]
    assert isinstance(emptyResult, ttb.sptensor)
    assert emptyResult.isequal(
        ttb.sptensor.from_data(np.array([]), np.array([]), (1, 1, 1))
    )
    # Full subtensor
    assert isinstance(sptensorInstance[:, :, :], ttb.sptensor)
    assert isinstance(
        sptensorInstance[[0, 1, 2, 3], [0, 1, 2, 3], [0, 1, 2, 3]], ttb.sptensor
    )
    assert sptensorInstance[:, :, :].isequal(
        sptensorInstance[[0, 1, 2, 3], [0, 1, 2, 3], [0, 1, 2, 3]]
    )

    # TODO need to understand what this intends to do
    ## Case 2 subscript indexing
    assert sptensorInstance[np.array([[1, 2, 1]])] == np.array([[0]])
    assert (
        sptensorInstance[np.array([[1, 2, 1], [1, 3, 1]])] == np.array([[0], [0]])
    ).all()

    ## Case 2 Linear Indexing
    ind = ttb.tt_sub2ind(data["shape"], np.array([[1, 1, 1], [1, 1, 3], [2, 2, 2]]))
    assert (sptensorInstance[ind] == np.array([[0.5], [1.5], [2.5]])).all()
    list_ind = list(ind)
    assert (sptensorInstance[list_ind] == np.array([[0.5], [1.5], [2.5]])).all()
    ind2 = ttb.tt_sub2ind(data["shape"], np.array([[1, 1, 1], [1, 1, 3]]))
    assert (sptensorInstance[ind2] == np.array([[0.5], [1.5]])).all()
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance[ind2[:, None]]
    assert "Expecting a row index" in str(excinfo)
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance["string"]
    assert "Invalid indexing" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor_setitem_Case1(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Empty sptensor assigned with nothing
    emptyTensor = ttb.sptensor()
    emptyTensor[:, :, :] = []
    assert emptyTensor.vals.size == 0
    emptyTensor[:, :, :] = np.array([])
    assert emptyTensor.vals.size == 0

    # Case I(a): Set empty tensor with sptensor
    emptyTensor = ttb.sptensor()
    emptyTensor[:, :, :] = sptensorInstance
    assert (emptyTensor.subs == data["subs"]).all()
    assert (emptyTensor.vals == data["vals"]).all()
    assert emptyTensor.shape == data["shape"]

    # Case I(a): Set empty tensor with sptensor, none none end slice
    emptyTensor = ttb.sptensor()
    emptyTensor[0:4, 0:4, 0:4] = sptensorInstance
    assert (emptyTensor.subs == data["subs"]).all()
    assert (emptyTensor.vals == data["vals"]).all()
    assert emptyTensor.shape == data["shape"]

    # Case I(a): Set sptensor with empty tensor
    emptyTensor = ttb.sptensor()
    emptyTensor.shape = (4, 4, 4)
    sptensorCopy = ttb.sptensor.from_tensor_type(sptensorInstance)
    sptensorCopy[:, :, :] = emptyTensor
    assert (sptensorCopy.subs == emptyTensor.subs).all()
    assert (sptensorCopy.vals == emptyTensor.vals).all()
    assert sptensorCopy.shape == data["shape"]

    # Case I(a): Set sptensor with smaller tensor
    emptyTensor = ttb.sptensor()
    emptyTensor.shape = (4, 4, 4)
    sptensorCopy = ttb.sptensor.from_tensor_type(sptensorInstance)
    sptensorInstanceCopy = ttb.sptensor.from_tensor_type(sptensorInstance)
    sptensorInstanceCopy[1, 1, 1] = 0
    sptensorCopy[4, 4, 4] = 1
    sptensorCopy[:4, :4, :4] = sptensorInstanceCopy
    assert (sptensorCopy.subs[1:, :] == sptensorInstanceCopy.subs).all()
    assert (sptensorCopy.vals[1:] == sptensorInstanceCopy.vals).all()
    assert sptensorCopy.shape == (5, 5, 5)

    # Case I(a): Set sptensor with smaller tensor
    emptyTensor = ttb.sptensor()
    emptyTensor.shape = (4, 4, 4)
    sptensorCopy = ttb.sptensor.from_tensor_type(sptensorInstance)
    sptensorCopy[4, 4, 4] = 1
    sptensorCopy[:4, :4, :4] = emptyTensor
    assert sptensorCopy.subs[1:, :].size == 0
    assert sptensorCopy.vals[1:].size == 0
    assert sptensorCopy.shape == (5, 5, 5)

    # Case I(a): Set sptensor with larger empty tensor
    emptyTensor = ttb.sptensor.from_data(np.array([]), np.array([]), (4, 4, 4, 4))
    sptensorCopy = ttb.sptensor.from_tensor_type(sptensorInstance)
    sptensorCopy[:4, :4, :4, :4] = emptyTensor
    assert (sptensorCopy.subs == emptyTensor.subs).all()
    assert (sptensorCopy.vals == emptyTensor.vals).all()
    assert sptensorCopy.shape == emptyTensor.shape

    # Case I(a): Set sptensor with sptensor
    subs = np.array([[2, 1, 1], [2, 1, 3]])
    vals = np.array([[2.5], [3.5]])
    shape = (4, 4, 4)
    sptensorRHS = ttb.sptensor.from_data(subs, vals, shape)

    sptensorCopy = ttb.sptensor.from_tensor_type(sptensorInstance)
    # TODO slicing in this way isn't supported in tensor toolbox
    # sptensorCopy[2, 1, :] = sptensorRHS
    # assert (sptensorCopy.subs == np.vstack((sptensorCopy.subs, sptensorRHS.subs))).all()
    # assert (sptensorCopy.vals == np.vstack((sptensorCopy.vals, sptensorRHS.vals))).all()
    # assert (sptensorCopy.shape == data['shape'])

    # Case I(a): Set empty with same size sptensor
    emptyTensor = ttb.sptensor.from_data(np.array([]), np.array([]), (1, 1, 1))
    sptensorCopy = ttb.sptensor.from_tensor_type(emptyTensor)
    sptensorCopy[0, 0, 0] = 1
    emptyTensor[0, 0, 0] = sptensorCopy
    # TODO: This ne should be eq once irenumber is resolved
    assert sptensorCopy.subs.shape != emptyTensor.subs.shape
    assert sptensorCopy.vals == emptyTensor.vals
    assert sptensorCopy.shape == emptyTensor.shape

    # Case I(a): Set empty with same size sptensor
    emptyTensor = ttb.sptensor.from_data(np.array([]), np.array([]), (1, 1, 1))
    sptensorCopy = ttb.sptensor.from_data(np.array([]), np.array([]), (1, 1, 1, 1))
    sptensorCopy[0, 0, 0, 0] = 1
    emptyTensor[0, 0, 0, 0] = sptensorCopy
    # TODO: This ne should be eq once irenumber is resolved
    assert sptensorCopy.subs.shape != emptyTensor.subs.shape
    assert sptensorCopy.vals == emptyTensor.vals
    # Since we do a single index set item the size is only set large enough for that element
    assert sptensorCopy.shape == emptyTensor.shape

    # Case I(a): Set empty with same size sptensor
    emptyTensor = ttb.sptensor.from_data(np.array([]), np.array([]), (2, 2, 2))
    sptensorCopy = ttb.sptensor.from_data(np.array([]), np.array([]), (2, 2, 2, 2))
    sptensorCopy[0, 0, 0, 0] = 1
    sptensorCopy[1, 1, 1, 1] = 1
    emptyTensor[[0, 1], [0, 1], [0, 1], [0, 1]] = sptensorCopy
    # TODO: This ne should be eq once irenumber is resolved
    assert (sptensorCopy.subs == emptyTensor.subs).all()
    assert (sptensorCopy.vals == emptyTensor.vals).all()
    assert sptensorCopy.shape == emptyTensor.shape

    sptensorCopy = ttb.sptensor.from_data(np.array([]), np.array([]), (1, 1, 1, 1))
    with pytest.raises(AssertionError) as excinfo:
        emptyTensor[[0, 1], [0, 1], [0, 1], [0, 1]] = sptensorCopy
    assert "RHS does not match range size" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        emptyTensor[
            np.array([0, 1]), np.array([0, 1]), np.array([0, 1]), np.array([0, 1])
        ] = sptensorCopy
    assert "RHS does not match range size" in str(excinfo)

    # Case I(b)i: Set with zero, sub already exists
    old_value = data["vals"][1, 0]
    sptensorInstance[1, 1, 3] = 0
    subSelection = [0, 2, 3]
    assert (sptensorInstance.subs == data["subs"][subSelection]).all()
    assert (sptensorInstance.vals == data["vals"][subSelection]).all()
    assert sptensorInstance.shape == data["shape"]

    # Case I(b)i: Set with zero, sub doesn't exist
    sptensorInstance[1, 1, 3] = old_value
    reorder = [0, 2, 3, 1]
    assert (sptensorInstance.subs == data["subs"][reorder]).all()
    assert (sptensorInstance.vals == data["vals"][reorder]).all()
    assert sptensorInstance.shape == data["shape"]
    # Reset tensor data
    data["subs"] = data["subs"][reorder]
    data["vals"] = data["vals"][reorder]

    # Case I(b)i: Set slice with zero, sub already exists
    old_value = data["vals"][3, 0]
    sptensorInstance[1:2, 1:2, 3:4] = 0
    subSelection = [0, 1, 2]
    assert (sptensorInstance.subs == data["subs"][subSelection]).all()
    assert (sptensorInstance.vals == data["vals"][subSelection]).all()
    assert sptensorInstance.shape == data["shape"]
    # Reset tensor data
    sptensorInstance[1, 1, 3] = old_value

    # Case I(b)i: Set slice with zero, sub already exists
    old_value = data["vals"][2, 0]
    sptensorInstance[3:, 3:, 3:] = 0
    subSelection = [0, 1, 3]
    assert (sptensorInstance.subs == data["subs"][subSelection]).all()
    assert (sptensorInstance.vals == data["vals"][subSelection]).all()
    assert sptensorInstance.shape == data["shape"]
    # Reset tensor data
    sptensorInstance[3, 3, 3] = old_value
    reorder = [0, 1, 3, 2]
    # Reset tensor data
    data["subs"] = data["subs"][reorder]
    data["vals"] = data["vals"][reorder]

    # Case I(b)i: Expand Shape of sptensor with set item
    sptensorInstanceLarger = ttb.sptensor.from_tensor_type(sptensorInstance)
    sptensorInstanceLarger[1, 1, 1, 1] = 0
    assert sptensorInstanceLarger.shape == (4, 4, 4, 2)

    # Case I(b)i: Expand Shape of sptensor with set item
    sptensorInstanceLarger = ttb.sptensor.from_tensor_type(sptensorInstance)
    sptensorInstanceLarger[1, 1, 1, 1:2] = 0
    assert sptensorInstanceLarger.shape == (4, 4, 4, 2)

    # Case I(b)i: Expand Shape of sptensor with set item
    sptensorInstanceLarger = ttb.sptensor.from_tensor_type(sptensorInstance)
    sptensorInstanceLarger[1, 1, 1, np.array([1])] = 0
    assert sptensorInstanceLarger.shape == (4, 4, 4, 2)

    # Case I(b)i: Expand Shape of sptensor with set item
    sptensorInstanceLarger = ttb.sptensor.from_tensor_type(sptensorInstance)
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstanceLarger[1, 1, 1, 1:] = 0
    assert (
        "Must have well defined slice when expanding sptensor shape with setitem"
        in str(excinfo)
    )

    # Case I(b)ii: Set with scalar, sub already exists
    old_value = data["vals"][2, 0]
    sptensorInstance[1, 1, 3] = 7
    modifiedVals = data["vals"].copy()
    modifiedVals[2] = 7
    assert (sptensorInstance.subs == data["subs"]).all()
    assert (sptensorInstance.vals == modifiedVals).all()
    assert sptensorInstance.shape == data["shape"]
    sptensorInstance[1, 1, 3] = old_value  # Reset tensor

    # Case I(b)ii: Set with scalar, sub already exists
    old_value = data["vals"][2, 0]
    sptensorInstance[1:2, 1:2, 3:4] = 7
    modifiedVals = data["vals"].copy()
    modifiedVals[2] = 7
    assert (sptensorInstance.subs == data["subs"]).all()
    assert (sptensorInstance.vals == modifiedVals).all()
    assert sptensorInstance.shape == data["shape"]
    sptensorInstance[1, 1, 3] = old_value  # Reset tensor

    # Case I(b)ii: Set with scalar, sub doesn't exist yet
    sptensorInstance[1, 1, 2] = 7
    assert (
        sptensorInstance.subs == np.vstack((data["subs"], np.array([[1, 1, 2]])))
    ).all()
    assert (sptensorInstance.vals == np.vstack((data["vals"], np.array([[7]])))).all()
    assert sptensorInstance.shape == data["shape"]

    # Case I: Assign with non-scalar or sptensor
    sptensorInstanceLarger = ttb.sptensor.from_tensor_type(sptensorInstance)
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstanceLarger[1, 1, 1] = "String"
    assert "Invalid assignment value" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor_setitem_Case2(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Case II: Too few modes in setitem key
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance[np.array([1, 1]).astype(int)] = 999.0
    assert "Invalid subscripts" in str(excinfo)

    # Case II: Too few keys in setitem for number of assignement values
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance[np.array([1, 1, 1]).astype(int)] = np.array([[999.0], [888.0]])
    assert "Number of subscripts and number of values do not match!" in str(excinfo)

    # Case II: Warning For duplicates
    with pytest.warns(Warning) as record:
        sptensorInstance[np.array([[1, 1, 1], [1, 1, 1]]).astype(int)] = np.array(
            [[999.0], [999.0]]
        )
    assert "Duplicate assignments discarded" in str(record[0].message)

    # Case II: Single entry, for single sub that exists
    sptensorInstance[np.array([1, 1, 1]).astype(int)] = 999.0
    assert (sptensorInstance[np.array([[1, 1, 1]])] == np.array([[999]])).all()
    assert (sptensorInstance.subs == data["subs"]).all()

    # Case II: Single entry, for multiple subs that exist
    (data, sptensorInstance) = sample_sptensor
    sptensorInstance[np.array([[1, 1, 1], [1, 1, 3]]).astype(int)] = 999.0
    assert (
        sptensorInstance[np.array([[1, 1, 1], [1, 1, 3]])] == np.array([[999], [999]])
    ).all()
    assert (sptensorInstance.subs == data["subs"]).all()

    # Case II: Multiple entries, for multiple subs that exist
    (data, sptensorInstance) = sample_sptensor
    sptensorInstance[np.array([[1, 1, 1], [1, 1, 3]]).astype(int)] = np.array(
        [[888], [999]]
    )
    assert (
        sptensorInstance[np.array([[1, 1, 3], [1, 1, 1]])] == np.array([[999], [888]])
    ).all()
    assert (sptensorInstance.subs == data["subs"]).all()

    # Case II: Single entry, for single sub that doesn't exist
    (data, sptensorInstance) = sample_sptensor
    copy = ttb.sptensor.from_tensor_type(sptensorInstance)
    copy[np.array([[1, 1, 2]]).astype(int)] = 999.0
    assert (copy[np.array([[1, 1, 2]])] == np.array([999])).all()
    assert (copy.subs == np.concatenate((data["subs"], np.array([[1, 1, 2]])))).all()

    # Case II: Single entry, for single sub that doesn't exist, expand dimensions
    (data, sptensorInstance) = sample_sptensor
    copy = ttb.sptensor.from_tensor_type(sptensorInstance)
    copy[np.array([[1, 1, 2, 1]]).astype(int)] = 999.0
    assert (copy[np.array([[1, 1, 2, 1]])] == np.array([999])).all()
    # assert (copy.subs == np.concatenate((data['subs'], np.array([[1, 1, 2]])))).all()

    # Case II: Single entry, for multiple subs one that exists and the other doesn't
    (data, sptensorInstance) = sample_sptensor
    copy = ttb.sptensor.from_tensor_type(sptensorInstance)
    copy[np.array([[1, 1, 1], [2, 1, 3]]).astype(int)] = 999.0
    assert (copy[np.array([[2, 1, 3]])] == np.array([999])).all()
    assert (copy.subs == np.concatenate((data["subs"], np.array([[2, 1, 3]])))).all()

    # Case II: Multiple entries, for multiple subs that don't exist
    (data, sptensorInstance) = sample_sptensor
    copy = ttb.sptensor.from_tensor_type(sptensorInstance)
    copy[np.array([[1, 1, 2], [2, 1, 3]]).astype(int)] = np.array([[888], [999]])
    assert (copy[np.array([[1, 1, 2], [2, 1, 3]])] == np.array([[888], [999]])).all()
    assert (
        copy.subs == np.concatenate((data["subs"], np.array([[1, 1, 2], [2, 1, 3]])))
    ).all()

    # Case II: Multiple entries, for multiple subs that exist and need to be removed
    (data, sptensorInstance) = sample_sptensor
    copy = ttb.sptensor.from_tensor_type(sptensorInstance)
    copy[np.array([[1, 1, 1], [1, 1, 3]]).astype(int)] = np.array([[0], [0]])
    assert (copy[np.array([[1, 1, 2], [2, 1, 3]])] == np.array([[0], [0]])).all()
    assert (copy.subs == np.array([[2, 2, 2], [3, 3, 3]])).all()


@pytest.mark.indevelopment
def test_sptensor_norm(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    assert sptensorInstance.norm() == np.linalg.norm(data["vals"])


@pytest.mark.indevelopment
def test_sptensor_allsubs(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    result = []
    for i in range(0, data["shape"][0]):
        for j in range(0, data["shape"][1]):
            for k in range(0, data["shape"][2]):
                result.append([i, j, k])
    assert (sptensorInstance.allsubs() == np.array(result)).all()


@pytest.mark.indevelopment
def test_sptensor_logical_not(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    result = []
    data_subs = data["subs"].tolist()
    for i in range(0, data["shape"][0]):
        for j in range(0, data["shape"][1]):
            for k in range(0, data["shape"][2]):
                if [i, j, k] not in data_subs:
                    result.append([i, j, k])
    notSptensorInstance = sptensorInstance.logical_not()
    assert (notSptensorInstance.vals == 1).all()
    assert (notSptensorInstance.subs == np.array(result)).all()
    assert notSptensorInstance.shape == data["shape"]


@pytest.mark.indevelopment
def test_sptensor_logical_or(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Sptensor logical or with another sptensor
    sptensorOr = sptensorInstance.logical_or(sptensorInstance)
    assert sptensorOr.shape == data["shape"]
    assert (sptensorOr.subs == data["subs"]).all()
    assert (sptensorOr.vals == np.ones((data["vals"].shape[0], 1))).all()

    # Sptensor logical or with tensor
    sptensorOr = sptensorInstance.logical_or(
        ttb.tensor.from_tensor_type(sptensorInstance)
    )
    nonZeroMatrix = np.zeros(data["shape"])
    nonZeroMatrix[tuple(data["subs"].transpose())] = 1
    assert (sptensorOr.data == nonZeroMatrix).all()

    # Sptensor logical or with scalar, 0
    sptensorOr = sptensorInstance.logical_or(0)
    assert (sptensorOr.data == nonZeroMatrix).all()

    # Sptensor logical or with scalar, not 0
    sptensorOr = sptensorInstance.logical_or(1)
    assert (sptensorOr.data == np.ones(data["shape"])).all()

    # Sptensor logical or with wrong shape sptensor
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.logical_or(
            ttb.sptensor.from_data(data["subs"], data["vals"], (5, 5, 5))
        )
    assert "Logical Or requires tensors of the same size" in str(excinfo)

    # Sptensor logical or with not scalar or tensor
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.logical_or(np.ones(data["shape"]))
    assert "Sptensor Logical Or argument must be scalar or sptensor" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor__eq__(sample_sptensor):
    # TODO fix == against empty sptensor
    (data, sptensorInstance) = sample_sptensor

    eqSptensor = sptensorInstance == 0.0
    assert (eqSptensor.subs == sptensorInstance.logical_not().subs).all()

    eqSptensor = sptensorInstance == 0.5
    assert (eqSptensor.subs == data["subs"][0]).all()

    eqSptensor = sptensorInstance == sptensorInstance
    assert (
        eqSptensor.subs
        == np.vstack((sptensorInstance.logical_not().subs, data["subs"]))
    ).all()

    denseTensor = ttb.tensor.from_tensor_type(sptensorInstance)
    eqSptensor = sptensorInstance == denseTensor
    logging.debug(f"\ndenseTensor = {denseTensor}")
    logging.debug(f"\nsptensorInstance = {sptensorInstance}")
    logging.debug(f"\ntype(eqSptensor.subs) = \n{type(eqSptensor.subs)}")
    for i in range(eqSptensor.subs.shape[0]):
        logging.debug(f"{i}\t{eqSptensor.subs[i,:]}")
    logging.debug(f"\neqSptensor.subs = \n{eqSptensor.subs}")
    logging.debug(f"\neqSptensor.subs.shape[0] = {eqSptensor.subs.shape[0]}")
    logging.debug(f"\nsptensorInstance.shape = {sptensorInstance.shape}")
    logging.debug(
        f"\nnp.prod(sptensorInstance.shape) = {np.prod(sptensorInstance.shape)}"
    )
    assert eqSptensor.subs.shape[0] == np.prod(sptensorInstance.shape)

    denseTensor = ttb.tensor.from_data(np.ones((5, 5, 5)))
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance == denseTensor
    assert "Size mismatch in sptensor equality" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance == np.ones((4, 4, 4))
    assert "Sptensor == argument must be scalar or sptensor" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor__ne__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    X = ttb.sptensor.from_data(np.array([[0, 0], [1, 1]]), np.array([[2], [2]]), (2, 2))
    Y = ttb.sptensor.from_data(np.array([[0, 0], [0, 1]]), np.array([[3], [3]]), (2, 2))
    assert (X != Y).isequal(
        ttb.sptensor.from_data(
            np.array([[1, 1], [0, 1], [0, 0]]),
            np.array([True, True, True])[:, None],
            (2, 2),
        )
    )

    eqSptensor = sptensorInstance != 0.0
    assert (eqSptensor.vals == 0 * sptensorInstance.vals + 1).all()

    eqSptensor = sptensorInstance != 0.5
    assert (
        eqSptensor.subs
        == np.vstack((data["subs"][1:], sptensorInstance.logical_not().subs))
    ).all()

    eqSptensor = sptensorInstance != sptensorInstance
    assert eqSptensor.vals.size == 0

    denseTensor = ttb.tensor.from_tensor_type(sptensorInstance)
    eqSptensor = sptensorInstance != denseTensor
    assert eqSptensor.vals.size == 0

    denseTensor = ttb.tensor.from_tensor_type(sptensorInstance)
    denseTensor[1, 1, 2] = 1
    eqSptensor = sptensorInstance != denseTensor
    assert (eqSptensor.subs == np.array([1, 1, 2])).all()

    denseTensor = ttb.tensor.from_data(np.ones((5, 5, 5)))
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance != denseTensor
    assert "Size mismatch" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance != np.ones((4, 4, 4))
    assert "The arguments must be two sptensors or an sptensor and a scalar." in str(
        excinfo
    )


def test_sptensor__end(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    assert sptensorInstance.end() == np.prod(data["shape"]) - 1
    assert sptensorInstance.end(k=0) == data["shape"][0] - 1


def test_sptensor__find(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    subs, vals = sptensorInstance.find()
    assert (subs == data["subs"]).all()
    assert (vals == data["vals"]).all()


def test_sptensor__sub__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Sptensor - sptensor
    subSptensor = sptensorInstance - sptensorInstance
    assert subSptensor.vals.size == 0

    # Sptensor - sptensor of wrong size
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance - ttb.sptensor.from_data(np.array([]), np.array([]), (6, 6, 6))
    assert "Must be two sparse tensors of the same shape" in str(excinfo)

    # Sptensor - tensor
    subSptensor = sptensorInstance - ttb.tensor.from_tensor_type(sptensorInstance)
    assert (subSptensor.data == np.zeros(data["shape"])).all()

    # Sptensor - scalar
    subSptensor = sptensorInstance - 0
    assert (
        subSptensor.data == ttb.tensor.from_tensor_type(sptensorInstance).data
    ).all()


def test_sptensor__add__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Sptensor + sptensor
    subSptensor = sptensorInstance + sptensorInstance
    assert (subSptensor.vals == 2 * data["vals"]).all()

    # Sptensor + sptensor of wrong size
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance + ttb.sptensor.from_data(np.array([]), np.array([]), (6, 6, 6))
    assert "Must be two sparse tensors of the same shape" in str(excinfo)

    # Sptensor + tensor
    subSptensor = sptensorInstance + ttb.tensor.from_tensor_type(sptensorInstance)
    results = ttb.tensor.from_tensor_type(sptensorInstance).data * 2
    assert (subSptensor.data == results).all()

    # Sptensor + scalar
    subSptensor = sptensorInstance + 0
    assert (
        subSptensor.data == ttb.tensor.from_tensor_type(sptensorInstance).data
    ).all()


def test_sptensor_isequal(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Wrong shape sptensor
    assert not sptensorInstance.isequal(
        ttb.sptensor.from_data(np.array([]), np.array([]), (6, 6, 6))
    )

    # Sptensor is equal to itself
    assert sptensorInstance.isequal(sptensorInstance)

    # Sptensor equality with tensor
    assert sptensorInstance.isequal(ttb.tensor.from_tensor_type(sptensorInstance))

    # Sptensor equality with not sptensor or tensor
    assert not sptensorInstance.isequal(np.ones(data["shape"]))


def test_sptensor__pos__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    sptensorInstance2 = +sptensorInstance

    assert sptensorInstance.isequal(sptensorInstance2)


def test_sptensor__neg__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    sptensorInstance2 = -sptensorInstance
    sptensorInstance3 = -sptensorInstance2

    assert not sptensorInstance.isequal(sptensorInstance2)
    assert sptensorInstance.isequal(sptensorInstance3)


def test_sptensor__mul__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Test mul with int
    assert ((sptensorInstance * 2).vals == 2 * data["vals"]).all()
    # Test mul with float
    assert ((sptensorInstance * 2.0).vals == 2 * data["vals"]).all()
    # Test mul with sptensor
    assert (
        (sptensorInstance * sptensorInstance).vals == data["vals"] * data["vals"]
    ).all()
    # Test mul with tensor
    assert (
        (sptensorInstance * ttb.tensor.from_tensor_type(sptensorInstance)).vals
        == data["vals"] * data["vals"]
    ).all()
    # Test mul with ktensor
    weights = np.array([1.0, 2.0])
    fm0 = np.array([[1.0, 2.0], [3.0, 4.0]])
    fm1 = np.array([[5.0, 6.0], [7.0, 8.0]])
    factor_matrices = [fm0, fm1]
    K = ttb.ktensor.from_data(weights, factor_matrices)
    subs = np.array([[0, 0], [0, 1], [1, 1]])
    vals = np.array([[0.5], [1.0], [1.5]])
    shape = (2, 2)
    S = ttb.sptensor().from_data(subs, vals, shape)
    assert (S * K).full().isequal(K.full() * S)

    # Test mul with wrong shape
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance * ttb.sptensor.from_data(np.array([]), np.array([]), (5, 5, 5))
    assert "Sptensor Multiply requires two tensors of the same shape." in str(excinfo)

    # Test mul with wrong type
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance * "string"
    assert "Sptensor cannot be multiplied by that type of object" in str(excinfo)


def test_sptensor__rmul__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Test mul with int
    assert ((2 * sptensorInstance).vals == 2 * data["vals"]).all()
    # Test mul with float
    assert ((2.0 * sptensorInstance).vals == 2 * data["vals"]).all()
    # Test mul with ktensor
    weights = np.array([1.0, 2.0])
    fm0 = np.array([[1.0, 2.0], [3.0, 4.0]])
    fm1 = np.array([[5.0, 6.0], [7.0, 8.0]])
    factor_matrices = [fm0, fm1]
    K = ttb.ktensor.from_data(weights, factor_matrices)
    subs = np.array([[0, 0], [0, 1], [1, 1]])
    vals = np.array([[0.5], [1.0], [1.5]])
    shape = (2, 2)
    S = ttb.sptensor().from_data(subs, vals, shape)
    assert (S * K).full().isequal(S * K.full())

    # Test mul with wrong type
    with pytest.raises(AssertionError) as excinfo:
        "string" * sptensorInstance
    assert "This object cannot be multiplied by sptensor" in str(excinfo)


def test_sptensor_ones(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    assert (sptensorInstance.ones().vals == (0.0 * data["vals"] + 1)).all()


@pytest.mark.indevelopment
def test_sptensor_double(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    denseData = np.zeros(sptensorInstance.shape)
    actualIdx = tuple(data["subs"].transpose())
    denseData[actualIdx] = data["vals"].transpose()[0]

    assert (sptensorInstance.double() == denseData).all()
    assert sptensorInstance.double().shape == data["shape"]


@pytest.mark.indevelopment
def test_sptensor__le__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Test comparison to negative scalar
    assert ((-sptensorInstance <= -0.1).vals == 0 * data["vals"] + 1).all()
    # Test comparison to positive scalar
    assert ((sptensorInstance <= 0.1).vals == sptensorInstance.logical_not().vals).all()

    # Test comparison to tensor
    assert (
        (sptensorInstance <= sptensorInstance.full()).vals
        == np.ones((np.prod(data["shape"]), 1))
    ).all()

    # Test comparison to sptensor
    assert (
        (sptensorInstance <= sptensorInstance).vals
        == np.ones((np.prod(data["shape"]), 1))
    ).all()

    # Test comparison of empty tensor with sptensor, both ways
    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), data["shape"])
    assert (
        (emptySptensor <= sptensorInstance).vals == np.ones((np.prod(data["shape"]), 1))
    ).all()
    assert (
        (sptensorInstance <= emptySptensor).vals == sptensorInstance.logical_not().vals
    ).all()

    # Test comparison with different size
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance <= ttb.sptensor.from_data(
            np.array([]), np.array([]), (5, 5, 5)
        )
    assert "Size mismatch" in str(excinfo)

    # Test comparison with incorrect type
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance <= "string"
    assert "Cannot compare sptensor with that type" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor__ge__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Test comparison to positive scalar
    assert ((sptensorInstance >= 0.1).vals == 0 * data["vals"] + 1).all()
    # Test comparison to negative scalar
    assert (
        (sptensorInstance >= -0.1).vals == np.ones((np.prod(data["shape"]), 1))
    ).all()

    # Test comparison to tensor
    assert (
        (sptensorInstance >= sptensorInstance.full()).vals
        == np.ones((np.prod(data["shape"]), 1))
    ).all()

    # Test comparison to sptensor
    assert (
        (sptensorInstance >= sptensorInstance).vals
        == np.ones((np.prod(data["shape"]), 1))
    ).all()

    # Test comparison with different size
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance >= ttb.sptensor.from_data(
            np.array([]), np.array([]), (5, 5, 5)
        )
    assert "Size mismatch" in str(excinfo)

    # Test comparison with incorrect type
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance >= "string"
    assert "Cannot compare sptensor with that type" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor__gt__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Test comparison to positive scalar
    assert ((sptensorInstance > 0.1).vals == 0 * data["vals"] + 1).all()
    # Test comparison to negative scalar
    assert (
        (sptensorInstance > -0.1).vals == np.ones((np.prod(data["shape"]), 1))
    ).all()

    # Test comparison to tensor
    assert (sptensorInstance > sptensorInstance.full()).vals.size == 0

    # Test comparison to tensor of different sparsity patter
    denseTensor = sptensorInstance.full()
    denseTensor[1, 1, 2] = -1
    assert ((sptensorInstance > denseTensor).subs == np.array([1, 1, 2])).all()

    # Test comparison to sptensor
    assert (sptensorInstance > sptensorInstance).vals.size == 0

    # Test comparison with different size
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance > ttb.sptensor.from_data(np.array([]), np.array([]), (5, 5, 5))
    assert "Size mismatch" in str(excinfo)

    # Test comparison with incorrect type
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance > "string"
    assert "Cannot compare sptensor with that type" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor__lt__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Test comparison to negative scalar
    assert ((-sptensorInstance < -0.1).vals == 0 * data["vals"] + 1).all()
    # Test comparison to positive scalar
    assert ((sptensorInstance < 0.1).vals == sptensorInstance.logical_not().vals).all()

    # Test comparison to tensor
    assert (sptensorInstance < sptensorInstance.full()).vals.size == 0

    # Test comparison to sptensor
    assert (sptensorInstance < sptensorInstance).vals.size == 0

    # Test comparison of empty tensor with sptensor, both ways
    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), data["shape"])
    assert ((emptySptensor < sptensorInstance).subs == data["subs"]).all()
    assert (sptensorInstance < emptySptensor).vals.size == 0

    # Test comparison with different size
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance < ttb.sptensor.from_data(np.array([]), np.array([]), (5, 5, 5))
    assert "Size mismatch" in str(excinfo)

    # Test comparison with incorrect type
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance < "string"
    assert "Cannot compare sptensor with that type" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor_innerprod(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Empty sptensor innerproduct
    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), data["shape"])
    assert sptensorInstance.innerprod(emptySptensor) == 0
    assert emptySptensor.innerprod(sptensorInstance) == 0

    # Sptensor innerproduct
    assert sptensorInstance.innerprod(sptensorInstance) == data["vals"].transpose().dot(
        data["vals"]
    )

    # Sptensor innerproduct, other has more elements
    sptensorCopy = ttb.sptensor.from_tensor_type(sptensorInstance)
    sptensorCopy[0, 0, 0] = 1
    assert sptensorInstance.innerprod(sptensorCopy) == data["vals"].transpose().dot(
        data["vals"]
    )

    # Wrong shape sptensor
    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), (1, 1))
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.innerprod(emptySptensor)
    assert "Sptensors must be same shape for innerproduct" in str(excinfo)

    # Tensor innerproduct
    assert sptensorInstance.innerprod(
        ttb.tensor.from_tensor_type(sptensorInstance)
    ) == data["vals"].transpose().dot(data["vals"])

    # Wrong shape tensor
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.innerprod(ttb.tensor.from_data(np.array([1])))
    assert "Sptensor and tensor must be same shape for innerproduct" in str(excinfo)

    # Wrong type for innerprod
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.innerprod(5)
    assert f"Inner product between sptensor and {type(5)} not supported" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor_logical_xor(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    nonZeroMatrix = np.zeros(data["shape"])
    nonZeroMatrix[tuple(data["subs"].transpose())] = 1

    # Sptensor logical xor with scalar, 0
    sptensorXor = sptensorInstance.logical_xor(0)
    assert (sptensorXor.data == nonZeroMatrix).all()

    # Sptensor logical xor with scalar, not 0
    sptensorXor = sptensorInstance.logical_xor(1)
    assert (sptensorXor.data == sptensorInstance.logical_not().full().data).all()

    # Sptensor logical xor with another sptensor
    sptensorXor = sptensorInstance.logical_xor(sptensorInstance)
    assert sptensorXor.shape == data["shape"]
    assert sptensorXor.vals.size == 0

    # Sptensor logical xor with tensor
    sptensorXor = sptensorInstance.logical_xor(
        ttb.tensor.from_tensor_type(sptensorInstance)
    )
    assert (sptensorXor.data == np.zeros(data["shape"], dtype=bool)).all()

    # Sptensor logical xor with wrong shape sptensor
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.logical_xor(
            ttb.sptensor.from_data(data["subs"], data["vals"], (5, 5, 5))
        )
    assert "Logical XOR requires tensors of the same size" in str(excinfo)

    # Sptensor logical xor with not scalar or tensor
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.logical_xor(np.ones(data["shape"]))
    assert "The argument must be an sptensor, tensor or scalar" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor_squeeze(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # No singleton dimensions
    assert (sptensorInstance.squeeze().vals == data["vals"]).all()
    assert (sptensorInstance.squeeze().subs == data["subs"]).all()

    # All singleton dimensions
    assert (
        ttb.sptensor.from_data(
            np.array([[0, 0, 0]]), np.array([4]), (1, 1, 1)
        ).squeeze()
        == 4
    )

    # A singleton dimension
    assert np.array_equal(
        ttb.sptensor.from_data(np.array([[0, 0, 0]]), np.array([4]), (2, 2, 1))
        .squeeze()
        .subs,
        np.array([[0, 0]]),
    )
    assert (
        ttb.sptensor.from_data(np.array([[0, 0, 0]]), np.array([4]), (2, 2, 1))
        .squeeze()
        .vals
        == np.array([4])
    ).all()

    # Singleton dimension with empty sptensor
    assert ttb.sptensor.from_data(
        np.array([]), np.array([]), (2, 2, 1)
    ).squeeze().shape == (2, 2)


@pytest.mark.indevelopment
def test_sptensor_scale(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Scale with np array
    assert (
        sptensorInstance.scale(np.array([4, 4, 4, 4]), 1).vals == 4 * data["vals"]
    ).all()

    # Scale with sptensor
    assert (
        sptensorInstance.scale(sptensorInstance, np.arange(0, 3)).vals
        == data["vals"] ** 2
    ).all()

    # Scale with tensor
    assert (
        sptensorInstance.scale(
            ttb.tensor.from_tensor_type(sptensorInstance), np.arange(0, 3)
        ).vals
        == data["vals"] ** 2
    ).all()

    # Incorrect shape np array, sptensor and tensor
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.scale(np.array([4, 4, 4, 4, 4]), 1)
    assert "Size mismatch in scale" in str(excinfo)
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.scale(
            ttb.sptensor.from_data(np.array([]), np.array([]), (1, 1, 1, 1, 1)),
            np.arange(0, 3),
        )
    assert "Size mismatch in scale" in str(excinfo)
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.scale(
            ttb.tensor.from_data(np.ones((1, 1, 1, 1, 1))), np.arange(0, 3)
        )
    assert "Size mismatch in scale" in str(excinfo)

    # Scale with non nparray, sptensor or tensor
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.scale(1, 1)
    assert "Invalid scaling factor" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor_reshape(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Standard reshape
    assert sptensorInstance.reshape((16, 4, 1)).shape == (16, 4, 1)

    # Reshape first and last modes, leave middle alone
    assert sptensorInstance.reshape((16, 1), np.array([0, 2])).shape == (4, 16, 1)

    # Reshape empty sptensor
    assert ttb.sptensor.from_data(np.array([]), np.array([]), (4, 4, 4)).reshape(
        (16, 4, 1)
    ).shape == (16, 4, 1)

    # Improper reshape
    with pytest.raises(AssertionError) as excinfo:
        assert sptensorInstance.reshape((16, 1), np.array([0])).shape == (4, 16, 1)
    assert "Reshape must maintain tensor size" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor_mask(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Mask captures all non-zero entries
    assert (sptensorInstance.mask(sptensorInstance) == data["vals"]).all()

    # Mask too large
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.mask(
            ttb.sptensor.from_data(np.array([]), np.array([]), (3, 3, 5))
        )
    assert "Mask cannot be bigger than the data tensor" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor_permute(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), (4, 4, 4))
    assert emptySptensor.permute(np.arange(0, 3)).isequal(emptySptensor)

    assert sptensorInstance.permute(np.arange(0, 3)).isequal(sptensorInstance)

    # Permute with too many dimensions
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.permute(np.arange(0, 4))
    assert "Invalid permutation order" in str(excinfo)

    # Permute that doesn't consider each dimension
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.permute(np.array([0, 0, 0]))
    assert "Invalid permutation order" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor__rtruediv__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Scalar / Spensor yields tensor, only resolves when left object doesn't have appropriate __truediv__
    # We ignore the divide by zero errors because np.inf/np.nan is an appropriate representation
    with np.errstate(divide="ignore", invalid="ignore"):
        assert ((2 / sptensorInstance).data == (2 / sptensorInstance.full().data)).all()

    # Tensor / Spensor yields tensor should be calling tensor.__truediv__
    # We ignore the divide by zero errors because np.inf/np.nan is an appropriate representation
    with np.errstate(divide="ignore", invalid="ignore"):
        np.testing.assert_array_equal(
            (sptensorInstance.full() / sptensorInstance).data,
            (sptensorInstance.full().data / sptensorInstance.full().data),
        )

    # Non-Scalar / Spensor yields tensor, only resolves when left object doesn't have appropriate __truediv__
    with pytest.raises(AssertionError) as excinfo:
        (("string" / sptensorInstance).data == (2 / sptensorInstance.full().data))
    assert "Dividing that object by an sptensor is not supported" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor__truediv__(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), (4, 4, 4))

    # Sptensor/ non-zero scalar
    assert ((sptensorInstance / 5).vals == data["vals"] / 5).all()

    # Sptensor/zero scalar
    np.testing.assert_array_equal(
        (sptensorInstance / 0).vals,
        np.vstack(
            (
                np.inf * np.ones((data["subs"].shape[0], 1)),
                np.nan * np.ones((np.prod(data["shape"]) - data["subs"].shape[0], 1)),
            )
        ),
    )

    # Sptensor/sptensor
    np.testing.assert_array_equal(
        (sptensorInstance / sptensorInstance).vals,
        np.vstack(
            (
                data["vals"] / data["vals"],
                np.nan * np.ones((np.prod(data["shape"]) - data["subs"].shape[0], 1)),
            )
        ),
    )

    # Sptensor/ empty tensor
    np.testing.assert_array_equal(
        (sptensorInstance / emptySptensor).vals,
        np.nan * np.ones((np.prod(data["shape"]), 1)),
    )

    # empty tensor/Sptensor
    np.testing.assert_array_equal(
        (emptySptensor / sptensorInstance).vals,
        np.vstack(
            (
                np.zeros((data["subs"].shape[0], 1)),
                np.nan * np.ones((np.prod(data["shape"]) - data["subs"].shape[0], 1)),
            )
        ),
    )

    # Sptensor/tensor
    assert (
        (sptensorInstance / sptensorInstance.full()).vals == data["vals"] / data["vals"]
    ).all()

    # Sptensor/ktensor
    weights = np.array([1.0, 2.0])
    fm0 = np.array([[1.0, 2.0], [3.0, 4.0]])
    fm1 = np.array([[5.0, 6.0], [7.0, 8.0]])
    factor_matrices = [fm0, fm1]
    K = ttb.ktensor.from_data(weights, factor_matrices)
    subs = np.array([[0, 0], [0, 1], [1, 1]])
    vals = np.array([[0.5], [1.0], [1.5]])
    shape = (2, 2)
    S = ttb.sptensor().from_data(subs, vals, shape)
    assert (S / K).full().isequal(S.full() / K.full())

    # Sptensor/ invalid
    with pytest.raises(AssertionError) as excinfo:
        (sptensorInstance / "string")
    assert "Invalid arguments for sptensor division" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        emptySptensor.shape = (5, 5, 5)
        (sptensorInstance / emptySptensor)
    assert "Sptensor division requires tensors of the same shape" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor_collapse(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), (4, 4, 4))

    # Test with no arguments
    assert sptensorInstance.collapse() == np.sum(data["vals"])

    # Test with custom function
    assert sptensorInstance.collapse(fun=sum) == np.sum(data["vals"])

    # Test partial collapse, output vector
    assert (
        sptensorInstance.collapse(dims=np.array([0, 1])) == np.array([0, 0.5, 2.5, 5])
    ).all()
    assert (
        emptySptensor.collapse(dims=np.array([0, 1])) == np.array([0, 0, 0, 0])
    ).all()

    # Test partial collapse, output sptensor
    collapseSptensor = sptensorInstance.collapse(dims=np.array([0]))
    assert (collapseSptensor.vals == data["vals"]).all()
    assert collapseSptensor.shape == (4, 4)
    assert (collapseSptensor.subs == data["subs"][:, 1:3]).all()
    emptySptensorSmaller = ttb.sptensor.from_tensor_type(emptySptensor)
    emptySptensorSmaller.shape = (4, 4)
    assert emptySptensor.collapse(dims=np.array([0])).isequal(emptySptensorSmaller)


@pytest.mark.indevelopment
def test_sptensor_contract(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), (5, 4, 4))

    with pytest.raises(AssertionError) as excinfo:
        emptySptensor.contract(0, 1)
    assert "Must contract along equally sized dimensions" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        emptySptensor.contract(0, 0)
    assert "Must contract along two different dimensions" in str(excinfo)

    contractableSptensor = ttb.sptensor.from_tensor_type(sptensorInstance)
    contractableSptensor = contractableSptensor.collapse(np.array([0]))
    assert contractableSptensor.contract(0, 1) == 6.5

    contractableSptensor = ttb.sptensor.from_tensor_type(sptensorInstance)
    assert (
        contractableSptensor.contract(0, 1).data == np.array([0, 0.5, 2.5, 5])
    ).all()

    contractableSptensor = ttb.sptensor.from_tensor_type(sptensorInstance)
    contractableSptensor[3, 3, 3, 3] = 1
    assert contractableSptensor.contract(0, 1).shape == (4, 4)


@pytest.mark.indevelopment
def test_sptensor_elemfun(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    def plus1(y):
        return y + 1

    assert (sptensorInstance.elemfun(plus1).vals == 1 + data["vals"]).all()
    assert (sptensorInstance.elemfun(plus1).subs == data["subs"]).all()

    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), (4, 4, 4))
    assert emptySptensor.elemfun(plus1).vals.size == 0


@pytest.mark.indevelopment
def test_sptensor_spmatrix(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.spmatrix()
    assert "Sparse tensor must be two dimensional" in str(excinfo)

    # Test empty sptensor to empty sparse matrix
    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), (4, 4))
    a = emptySptensor.spmatrix()

    assert a.data.size == 0
    assert a.shape == emptySptensor.shape

    NonEmptySptensor = ttb.sptensor.from_data(
        np.array([[0, 0]]), np.array([[1]]), (4, 4)
    )
    fullData = np.zeros(NonEmptySptensor.shape)
    fullData[0, 0] = 1
    b = NonEmptySptensor.spmatrix()
    assert (b.toarray() == fullData).all()

    NonEmptySptensor = ttb.sptensor.from_data(
        np.array([[0, 1], [1, 0]]), np.array([[1], [2]]), (4, 4)
    )
    fullData = np.zeros(NonEmptySptensor.shape)
    fullData[0, 1] = 1
    fullData[1, 0] = 2
    b = NonEmptySptensor.spmatrix()
    assert (b.toarray() == fullData).all()

    NonEmptySptensor = ttb.sptensor.from_data(
        np.array([[0, 1], [2, 3]]), np.array([[1], [2]]), (4, 4)
    )
    fullData = np.zeros(NonEmptySptensor.shape)
    fullData[0, 1] = 1
    fullData[2, 3] = 2
    b = NonEmptySptensor.spmatrix()
    assert (b.toarray() == fullData).all()


@pytest.mark.indevelopment
def test_sptensor_ttv(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Returns single value
    onesSptensor = ttb.sptensor.from_tensor_type(ttb.tensor.from_data(np.ones((4, 4))))
    vector = np.array([1, 1, 1, 1])
    assert onesSptensor.ttv(np.array([vector, vector])) == 16

    # Wrong shape vector
    with pytest.raises(AssertionError) as excinfo:
        onesSptensor.ttv([vector, np.array([1, 2])])
    assert "Multiplicand is wrong size" in str(excinfo)

    # Returns vector shaped object
    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), (4, 4))
    onesSptensor = ttb.sptensor.from_tensor_type(ttb.tensor.from_data(np.ones((4, 4))))

    assert emptySptensor.ttv(vector, 0).isequal(
        ttb.sptensor.from_data(np.array([]), np.array([]), (4,))
    )
    assert onesSptensor.ttv(vector, 0).isequal(
        ttb.tensor.from_data(np.array([4, 4, 4, 4]))
    )
    emptySptensor[0, 0] = 1
    assert (emptySptensor.ttv(vector, 0).full().data == np.array([1, 0, 0, 0])).all()

    # Returns tensor shaped object
    emptySptensor = ttb.sptensor.from_data(np.array([]), np.array([]), (4, 4, 4))
    onesSptensor = ttb.sptensor.from_tensor_type(
        ttb.tensor.from_data(np.ones((4, 4, 4)))
    )
    assert emptySptensor.ttv(vector, 0).isequal(
        ttb.sptensor.from_data(np.array([]), np.array([]), (4, 4))
    )
    assert onesSptensor.ttv(vector, 0).isequal(
        ttb.tensor.from_data(4 * np.ones((4, 4)))
    )


@pytest.mark.indevelopment
def test_sptensor_mttkrp(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # MTTKRP with array of matrices
    # Note this is more of a regression test against the output of MATLAB TTB
    matrix = np.ones((4, 4))
    assert (
        sptensorInstance.mttkrp(np.array([matrix, matrix, matrix]), 0)
        == np.array(
            [[0, 0, 0, 0], [2, 2, 2, 2], [2.5, 2.5, 2.5, 2.5], [3.5, 3.5, 3.5, 3.5]]
        )
    ).all()
    assert (
        sptensorInstance.mttkrp(np.array([matrix, matrix, matrix]), 1)
        == sptensorInstance.mttkrp(np.array([matrix, matrix, matrix]), 0)
    ).all()
    assert (
        sptensorInstance.mttkrp(np.array([matrix, matrix, matrix]), 2)
        == np.array(
            [[0, 0, 0, 0], [0.5, 0.5, 0.5, 0.5], [2.5, 2.5, 2.5, 2.5], [5, 5, 5, 5]]
        )
    ).all()

    # MTTKRP with factor matrices from ktensor
    K = ttb.ktensor.from_factor_matrices([matrix, matrix, matrix])
    assert (
        sptensorInstance.mttkrp(np.array([matrix, matrix, matrix]), 0)
        == sptensorInstance.mttkrp(K, 0)
    ).all()
    assert (
        sptensorInstance.mttkrp(np.array([matrix, matrix, matrix]), 1)
        == sptensorInstance.mttkrp(K, 1)
    ).all()
    assert (
        sptensorInstance.mttkrp(np.array([matrix, matrix, matrix]), 2)
        == sptensorInstance.mttkrp(K, 2)
    ).all()

    # Wrong length input
    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.mttkrp(np.array([matrix, matrix, matrix, matrix]), 0)
    assert "List is the wrong length" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.mttkrp("string", 0)
    assert "Second argument must be ktensor or array" in str(excinfo)


@pytest.mark.indevelopment
def test_sptensor_nvecs(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor

    # Test for one eigenvector
    assert np.allclose((sptensorInstance.nvecs(1, 1)), np.array([0, 0, 0, 1])[:, None])
    assert np.allclose(
        (sptensorInstance.nvecs(1, 2)),
        np.array([[0, 0, 0, 1], [0, 0, 1, 0]]).transpose(),
    )

    # Test for r >= N-1, requires cast to dense
    with pytest.warns(Warning) as record:
        ans = np.zeros((4, 3))
        ans[3, 0] = 1
        ans[2, 1] = 1
        ans[1, 2] = 1
        assert np.allclose((sptensorInstance.nvecs(1, 3)), ans)
    assert (
        "Greater than or equal to sptensor.shape[n] - 1 eigenvectors requires cast to dense to solve"
        in str(record[0].message)
    )

    # Negative test, check for only singleton dims
    with pytest.raises(ValueError):
        single_val_sptensor = ttb.sptensor.from_data(
            np.array([[0, 0]]), np.array([1]), shape=(1, 1)
        )
        single_val_sptensor.nvecs(0, 0)


@pytest.mark.indevelopment
def test_sptensor_ttm(sample_sptensor):
    (data, sptensorInstance) = sample_sptensor
    result = np.zeros((4, 4, 4))
    result[:, 1, 1] = 0.5
    result[:, 1, 3] = 1.5
    result[:, 2, 2] = 2.5
    result[:, 3, 3] = 3.5
    result = ttb.tensor.from_data(result)
    result = ttb.sptensor.from_tensor_type(result)
    assert sptensorInstance.ttm(sparse.coo_matrix(np.ones((4, 4))), dims=0).isequal(
        result
    )
    assert sptensorInstance.ttm(
        sparse.coo_matrix(np.ones((4, 4))), dims=0, transpose=True
    ).isequal(result)

    # This is a multiway multiplication yielding a sparse tensor, yielding a dense tensor relies on tensor.ttm
    matrix = sparse.coo_matrix(np.eye(4))
    list_of_matrices = [matrix, matrix, matrix]
    assert sptensorInstance.ttm(list_of_matrices, dims=[0, 1, 2]).isequal(
        sptensorInstance
    )
    assert sptensorInstance.ttm(list_of_matrices, exclude_dims=2).isequal(
        sptensorInstance.ttm(list_of_matrices[0:-1], dims=[0, 1])
    )

    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.ttm(sparse.coo_matrix(np.ones((5, 5))), dims=0)
    assert "Matrix shape doesn't match tensor shape" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.ttm(np.array([1, 2, 3, 4]), dims=0)
    assert "Sptensor.ttm: second argument must be a matrix" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        sptensorInstance.ttm(sparse.coo_matrix(np.ones((4, 4))), dims=4)
    assert "dims must contain values in [0,self.dims)" in str(excinfo)

    sptensorInstance[0, :, :] = 1
    sptensorInstance[3, :, :] = 1
    result = np.zeros((4, 4, 4))
    result[0, :, :] = 4.0
    result[3, :, :] = 4.0
    result[1, :, 1] = 0.5
    result[1, :, 3] = 1.5
    result[2, :, 2] = 2.5

    # TODO: Ensure mode mappings are consistent between matlab and numpy
    # MATLAB is opposite orientation so the mapping from matlab to numpy is
    # {3:0, 2:2, 1:1}
    assert sptensorInstance.ttm(sparse.coo_matrix(np.ones((4, 4))), dims=1).isequal(
        ttb.tensor.from_data(result)
    )

    result = 2 * np.ones((4, 4, 4))
    result[:, 1, 1] = 2.5
    result[:, 1, 3] = 3.5
    result[:, 2, 2] = 4.5
    assert sptensorInstance.ttm(sparse.coo_matrix(np.ones((4, 4))), dims=0).isequal(
        ttb.tensor.from_data(result)
    )

    result = np.zeros((4, 4, 4))
    result[0, :, :] = 4.0
    result[3, :, :] = 4.0
    result[1, 1, :] = 2
    result[2, 2, :] = 2.5
    assert sptensorInstance.ttm(sparse.coo_matrix(np.ones((4, 4))), dims=2).isequal(
        ttb.tensor.from_data(result)
    )

    # Confirm reshape for non-square matrix
    assert sptensorInstance.ttm(sparse.coo_matrix(np.ones((1, 4))), dims=2).shape == (
        4,
        4,
        1,
    )


@pytest.mark.indevelopment
def test_sptensor_to_sparse_matrix():
    subs = np.array([[1, 1, 1], [1, 1, 3], [2, 2, 2], [3, 3, 3]])
    vals = np.array([[0.5], [1.5], [2.5], [3.5]])
    shape = (4, 4, 4)
    mode0 = sparse.coo_matrix(([0.5, 1.5, 2.5, 3.5], ([5, 13, 10, 15], [1, 1, 2, 3])))
    mode1 = sparse.coo_matrix(([0.5, 1.5, 2.5, 3.5], ([5, 13, 10, 15], [1, 1, 2, 3])))
    mode2 = sparse.coo_matrix(([0.5, 1.5, 2.5, 3.5], ([5, 5, 10, 15], [1, 3, 2, 3])))
    Ynt = [mode0, mode1, mode2]
    sptensorInstance = ttb.sptensor().from_data(subs, vals, shape)

    for mode in range(sptensorInstance.ndims):
        Xnt = tt_to_sparse_matrix(sptensorInstance, mode, True)
        assert (Xnt != Ynt[mode]).nnz == 0
        assert Xnt.shape == Ynt[mode].shape


@pytest.mark.indevelopment
def test_sptensor_from_sparse_matrix():
    subs = np.array([[1, 1, 1], [1, 1, 3], [2, 2, 2], [3, 3, 3]])
    vals = np.array([[0.5], [1.5], [2.5], [3.5]])
    shape = (4, 4, 4)
    sptensorInstance = ttb.sptensor().from_data(subs, vals, shape)
    for mode in range(sptensorInstance.ndims):
        sptensorCopy = ttb.sptensor.from_tensor_type(sptensorInstance)
        Xnt = tt_to_sparse_matrix(sptensorCopy, mode, True)
        Ynt = tt_from_sparse_matrix(Xnt, sptensorCopy.shape, mode, 0)
        assert sptensorCopy.isequal(Ynt)

    for mode in range(sptensorInstance.ndims):
        sptensorCopy = ttb.sptensor.from_tensor_type(sptensorInstance)
        Xnt = tt_to_sparse_matrix(sptensorCopy, mode, False)
        Ynt = tt_from_sparse_matrix(Xnt, sptensorCopy.shape, mode, 1)
        assert sptensorCopy.isequal(Ynt)



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import textwrap
import warnings

import numpy as np
import scipy

from pyttb import ktensor
from pyttb import pyttb_utils as ttb_utils
from pyttb import sptenmat, sptensor, tenmat, tensor

ALT_CORE_ERROR = "TTensor doesn't support non-tensor cores yet"


class ttensor(object):
    """
    TTENSOR Class for Tucker tensors (decomposed).

    """

    def __init__(self):
        """
        Create an empty decomposed tucker tensor

        Returns
        -------
        :class:`pyttb.ttensor`
        """
        # Empty constructor
        self.core = tensor()
        self.u = []

    @classmethod
    def from_data(cls, core, factors):
        """
        Construct an ttensor from fully defined core tensor and factor matrices.

        Parameters
        ----------
        core: :class: `ttb.tensor`
        factors: :class:`list(numpy.ndarray)`

        Returns
        -------
        :class:`pyttb.ttensor`

        Examples
        --------
        Import required modules:

        >>> import pyttb as ttb
        >>> import numpy as np

        Set up input data
        # Create ttensor with explicit data description

        >>> core_values = np.ones((2,2,2))
        >>> core = ttb.tensor.from_data(core_values)
        >>> factors = [np.ones((1,2))] * len(core_values.shape)
        >>> K0 = ttb.ttensor.from_data(core, factors)
        """
        ttensorInstance = ttensor()
        if isinstance(core, tensor):
            ttensorInstance.core = tensor.from_data(core.data, core.shape)
            ttensorInstance.u = factors.copy()
        else:
            # TODO support any tensor type with supported ops
            raise ValueError("TTENSOR doesn't yet support generic cores, only tensor")
        ttensorInstance._validate_ttensor()
        return ttensorInstance

    @classmethod
    def from_tensor_type(cls, source):
        """
        Converts other tensor types into a ttensor

        Parameters
        ----------
        source: :class:`pyttb.ttensor`

        Returns
        -------
        :class:`pyttb.ttensor`
        """
        # Copy Constructor
        if isinstance(source, ttensor):
            return cls.from_data(source.core, source.u)

    def _validate_ttensor(self):
        """
        Verifies the validity of constructed ttensor

        Returns
        -------
        """
        # Confirm all factors are matrices
        for factor_idx, factor in enumerate(self.u):
            if not isinstance(factor, np.ndarray):
                raise ValueError(
                    f"Factor matrices must be numpy arrays but factor {factor_idx} was {type(factor)}"
                )
            if len(factor.shape) != 2:
                raise ValueError(
                    f"Factor matrix {factor_idx} has shape {factor.shape} and is not a matrix!"
                )

        # Verify size consistency
        core_order = len(self.core.shape)
        num_matrices = len(self.u)
        if core_order != num_matrices:
            raise ValueError(
                f"CORE has order {core_order} but there are {num_matrices} factors"
            )
        for factor_idx, factor in enumerate(self.u):
            if factor.shape[-1] != self.core.shape[factor_idx]:
                raise ValueError(
                    f"Factor matrix {factor_idx} does not have {self.core.shape[factor_idx]} columns"
                )

    @property
    def shape(self):
        """
        Shape of the tensor this deconstruction represents.

        Returns
        -------
        tuple(int)
        """
        return tuple(factor.shape[0] for factor in self.u)

    def __repr__(self):  # pragma: no cover
        """
        String representation of a tucker tensor.

        Returns
        -------
        str
            Contains the core, and factor matrices as strings on different lines.
        """
        display_string = f"Tensor of shape: {self.shape}\n" f"\tCore is a "
        display_string += textwrap.indent(str(self.core), "\t")

        for factor_idx, factor in enumerate(self.u):
            display_string += f"\tU[{factor_idx}] = \n"
            display_string += textwrap.indent(str(factor), "\t\t")
            display_string += "\n"
        return display_string

    __str__ = __repr__

    def full(self):
        """
        Convert a ttensor to a (dense) tensor.

        Returns
        -------
        :class:`pyttb.tensor`
        """
        recomposed_tensor = self.core.ttm(self.u)

        # There is a small chance tensor could be sparse so ensure we cast that to dense.
        if not isinstance(recomposed_tensor, tensor):
            raise ValueError(ALT_CORE_ERROR)
        return recomposed_tensor

    def double(self):
        """
        Convert ttensor to an array of doubles

        Returns
        -------
        :class:`numpy.ndarray`
            copy of tensor data
        """
        return self.full().double()

    @property
    def ndims(self):
        """
        Number of dimensions of a ttensor.

        Returns
        -------
        int
            Number of dimensions of ttensor
        """
        return len(self.u)

    def isequal(self, other):
        """
        Component equality for ttensors

        Parameters
        ----------
        other: :class:`pyttb.ttensor`

        Returns
        -------
        bool: True if ttensors decompositions are identical, false otherwise
        """
        if not isinstance(other, ttensor):
            return False
        if self.ndims != other.ndims:
            return False
        return self.core.isequal(other.core) and all(
            np.array_equal(this_factor, other_factor)
            for this_factor, other_factor in zip(self.u, other.u)
        )

    def __pos__(self):
        """
        Unary plus (+) for ttensors. Does nothing.

        Returns
        -------
        :class:`pyttb.ttensor`, copy of tensor
        """

        return ttensor.from_tensor_type(self)

    def __neg__(self):
        """
        Unary minus (-) for ttensors

        Returns
        -------
        :class:`pyttb.ttensor`, copy of tensor
        """

        return ttensor.from_data(-self.core, self.u)

    def innerprod(self, other):
        """
        Efficient inner product with a ttensor

        Parameters
        ----------
        other: :class:`pyttb.tensor`, :class:`pyttb.sptensor`, :class:`pyttb.ktensor`,
        :class:`pyttb.ttensor`

        Returns
        -------
        float
        """
        if isinstance(other, ttensor):
            if self.shape != other.shape:
                raise ValueError(
                    "ttensors must have same shape to perform an innerproduct, but this ttensor "
                    f"has shape {self.shape} and the other has {other.shape}"
                )
            if np.prod(self.core.shape) > np.prod(other.core.shape):
                # Reverse arguments so the ttensor with the smaller core comes first.
                return other.innerprod(self)
            W = []
            for this_factor, other_factor in zip(self.u, other.u):
                W.append(this_factor.transpose().dot(other_factor))
            J = other.core.ttm(W)
            return self.core.innerprod(J)
        elif isinstance(other, (tensor, sptensor)):
            if self.shape != other.shape:
                raise ValueError(
                    "ttensors must have same shape to perform an innerproduct, but this ttensor "
                    f"has shape {self.shape} and the other has {other.shape}"
                )
            if np.prod(self.shape) < np.prod(self.core.shape):
                Z = self.full()
                return Z.innerprod(other)
            Z = other.ttm(self.u, transpose=True)
            return Z.innerprod(self.core)
        elif isinstance(other, ktensor):
            # Call ktensor implementation
            # TODO needs ttensor ttv
            return other.innerprod(self)
        else:
            raise ValueError(
                f"Inner product between ttensor and {type(other)} is not supported"
            )

    def __mul__(self, other):
        """
        Element wise multiplication (*) for ttensors (only scalars supported)

        Parameters
        ----------
        other: float, int

        Returns
        -------
        :class:`pyttb.ttensor`
        """
        if isinstance(other, (float, int, np.number)):
            return ttensor.from_data(self.core * other, self.u)
        raise ValueError(
            "This object cannot be multiplied by ttensor. Convert to full if trying to "
            "multiply ttensor by another tensor."
        )

    def __rmul__(self, other):
        """
        Element wise right multiplication (*) for ttensors (only scalars supported)

        Parameters
        ----------
        other: float, int

        Returns
        -------
        :class:`pyttb.ttensor`
        """
        if isinstance(other, (float, int, np.number)):
            return self.__mul__(other)
        raise ValueError("This object cannot be multiplied by ttensor")

    def ttv(self, vector, dims=None, exclude_dims=None):
        """
        TTensor times vector

        Parameters
        ----------
        vector: :class:`Numpy.ndarray`, list[:class:`Numpy.ndarray`]
        dims: :class:`Numpy.ndarray`, int
        """
        if dims is None and exclude_dims is None:
            dims = np.array([])
        # TODO make helper function to check scalar since re-used many places
        elif isinstance(dims, (float, int)):
            dims = np.array([dims])

        if isinstance(exclude_dims, (float, int)):
            exclude_dims = np.array([exclude_dims])

        # Check that vector is a list of vectors, if not place single vector as element in list
        if len(vector) > 0 and isinstance(vector[0], (int, float, np.int_, np.float_)):
            return self.ttv(np.array([vector]), dims, exclude_dims)

        # Get sorted dims and index for multiplicands
        dims, vidx = ttb_utils.tt_dimscheck(self.ndims, len(vector), dims, exclude_dims)

        # Check that each multiplicand is the right size.
        for i in range(dims.size):
            if vector[vidx[i]].shape != (self.shape[dims[i]],):
                raise ValueError("Multiplicand is wrong size")

        # Get remaining dimensions when we're done
        remdims = np.setdiff1d(np.arange(0, self.ndims), dims)

        # Create W to multiply with core, only populated remaining dims
        W = [None] * self.ndims
        for i in range(dims.size):
            dim_idx = dims[i]
            W[dim_idx] = self.u[dim_idx].transpose().dot(vector[vidx[i]])

        # Create new core
        newcore = self.core.ttv(W, dims)

        # Create final result
        if remdims.size == 0:
            return newcore
        else:
            return ttensor.from_data(newcore, [self.u[dim] for dim in remdims])

    def mttkrp(self, U, n):
        """
        Matricized tensor times Khatri-Rao product for ttensors.

        Parameters
        ----------
        U: array of matrices or ktensor
        n: multiplies by all modes except n

        Returns
        -------
        :class:`numpy.ndarray`
        """
        # NOTE: MATLAB version calculates an unused R here

        W = [None] * self.ndims
        for i in range(0, self.ndims):
            if i == n:
                continue
            W[i] = self.u[i].transpose().dot(U[i])

        Y = self.core.mttkrp(W, n)

        # Find each column of answer by multiplying by weights
        return self.u[n].dot(Y)

    def norm(self):
        """
        Compute the norm of a ttensor.
        Returns
        -------
        norm: float, Frobenius norm of Tensor
        """
        if np.prod(self.shape) > np.prod(self.core.shape):
            V = []
            for factor in self.u:
                V.append(factor.transpose().dot(factor))
            Y = self.core.ttm(V)
            tmp = Y.innerprod(self.core)
            return np.sqrt(tmp)
        else:
            return self.full().norm()

    def permute(self, order):
        """
        Permute dimensions for a ttensor

        Parameters
        ----------
        order: :class:`Numpy.ndarray`

        Returns
        -------
        :class:`pyttb.ttensor`
        """
        if not np.array_equal(np.arange(0, self.ndims), np.sort(order)):
            raise ValueError("Invalid permutation")
        new_core = self.core.permute(order)
        new_u = [self.u[idx] for idx in order]
        return ttensor.from_data(new_core, new_u)

    def ttm(self, matrix, dims=None, exclude_dims=None, transpose=False):
        """
        Tensor times matrix for ttensor

        Parameters
        ----------
        matrix: :class:`Numpy.ndarray`, list[:class:`Numpy.ndarray`]
        dims: :class:`Numpy.ndarray`, int
        transpose: bool
        """
        if dims is None and exclude_dims is None:
            dims = np.arange(self.ndims)
        elif isinstance(dims, list):
            dims = np.array(dims)
        elif np.isscalar(dims):
            if dims < 0:
                raise ValueError("Negative dims is currently unsupported, see #62")
            dims = np.array([dims])

        if isinstance(exclude_dims, (float, int)):
            exclude_dims = np.array([exclude_dims])

        if not isinstance(matrix, list):
            return self.ttm([matrix], dims, exclude_dims, transpose)

        # Check that the dimensions are valid
        dims, vidx = ttb_utils.tt_dimscheck(self.ndims, len(matrix), dims, exclude_dims)

        # Determine correct size index
        size_idx = int(not transpose)

        # Check that each multiplicand is the right size.
        for i in range(len(dims)):
            if matrix[vidx[i]].shape[size_idx] != self.shape[dims[i]]:
                raise ValueError(f"Multiplicand {i} is wrong size")

        # Do the actual multiplications in the specified modes.
        new_u = self.u.copy()
        for i in range(len(dims)):
            if transpose:
                new_u[dims[i]] = matrix[vidx[i]].transpose().dot(new_u[dims[i]])
            else:
                new_u[dims[i]] = matrix[vidx[i]].dot(new_u[dims[i]])

        return ttensor.from_data(self.core, new_u)

    def reconstruct(self, samples=None, modes=None):
        """
        Reconstruct or partially reconstruct tensor from ttensor.

        Parameters
        ----------
        samples: :class:`Numpy.ndarray`, list[:class:`Numpy.ndarray`]
        modes: :class:`Numpy.ndarray`, list[:class:`Numpy.ndarray`]

        Returns
        -------
        :class:`pyttb.ttensor`
        """
        # Default to sampling full tensor
        full_tensor_sampling = samples is None and modes is None
        if full_tensor_sampling:
            return self.full()

        if modes is None:
            modes = np.arange(self.ndims)
        elif isinstance(modes, list):
            modes = np.array(modes)
        elif np.isscalar(modes):
            modes = np.array([modes])

        if np.isscalar(samples):
            samples = [np.array([samples])]
        elif not isinstance(samples, list):
            samples = [samples]

        unequal_lengths = len(samples) != len(modes)
        if unequal_lengths:
            raise ValueError(
                "If samples and modes provides lengths must be equal, but "
                f"samples had length {len(samples)} and modes {len(modes)}"
            )

        full_samples = [np.array([])] * self.ndims
        for sample, mode in zip(samples, modes):
            if np.isscalar(sample):
                full_samples[mode] = np.array([sample])
            else:
                full_samples[mode] = sample

        shape = self.shape
        new_u = []
        for k in range(self.ndims):
            if len(full_samples[k]) == 0:
                # Skip empty samples
                new_u.append(self.u[k])
                continue
            elif (
                len(full_samples[k].shape) == 2
                and full_samples[k].shape[-1] == shape[k]
            ):
                new_u.append(full_samples[k].dot(self.u[k]))
            else:
                new_u.append(self.u[k][full_samples[k], :])

        return ttensor.from_data(self.core, new_u).full()

    def nvecs(self, n, r, flipsign=True):
        """
        Compute the leading mode-n vectors for a ttensor.

        Parameters
        ----------
        n: mode for tensor matricization
        r: number of eigenvalues
        flipsign: Make each column's largest element positive if true

        Returns
        -------
        :class:`numpy.ndarray`
        """
        # Compute inner product of all n-1 factors
        V = []
        for factor_idx, factor in enumerate(self.u):
            if factor_idx == n:
                V.append(factor)
            else:
                V.append(factor.transpose().dot(factor))
        H = self.core.ttm(V)

        if isinstance(H, sptensor):
            raise NotImplementedError(ALT_CORE_ERROR)
        else:
            HnT = tenmat.from_tensor_type(H.full(), cdims=np.array([n])).double()

        G = self.core

        if isinstance(G, sptensor):
            raise NotImplementedError(ALT_CORE_ERROR)
        else:
            GnT = tenmat.from_tensor_type(G.full(), cdims=np.array([n])).double()

        # Compute Xn * Xn'
        Y = HnT.transpose().dot(GnT.dot(self.u[n].transpose()))

        # TODO: Lifted from tensor, consider common location
        if r < Y.shape[0] - 1:
            w, v = scipy.sparse.linalg.eigsh(Y, r)
            v = v[:, (-np.abs(w)).argsort()]
            v = v[:, :r]
        else:
            warnings.warn(
                "Greater than or equal to tensor.shape[n] - 1 eigenvectors requires cast to dense to solve"
            )
            w, v = scipy.linalg.eigh(Y)
            v = v[:, (-np.abs(w)).argsort()]
            v = v[:, :r]

        if flipsign:
            idx = np.argmax(np.abs(v), axis=0)
            for i in range(v.shape[1]):
                if v[idx[i], i] < 0:
                    v[:, i] *= -1
        return v



import numpy as np
import pytest

import pyttb as ttb


@pytest.fixture()
def sample_tensor():
    data = np.array([[29, 39.0], [63.0, 85.0]])
    shape = (2, 2)
    params = {"data": data, "shape": shape}
    tensorInstance = ttb.tensor().from_data(data, shape)
    return params, tensorInstance


@pytest.fixture()
def sample_tensor_3way():
    shape = (3, 3, 3)
    data = np.array(range(1, 28)).reshape(shape, order="F")
    params = {"data": data, "shape": shape}
    tensorInstance = ttb.tensor().from_data(data, shape)
    return params, tensorInstance


@pytest.mark.indevelopment
def test_hosvd_simple_convergence(capsys, sample_tensor):
    (data, T) = sample_tensor
    tol = 1e-4
    result = ttb.hosvd(T, tol)
    assert (result.full() - T).norm() / T.norm() < tol, f"Failed to converge"

    tol = 1e-4
    result = ttb.hosvd(T, tol, sequential=False)
    assert (
        result.full() - T
    ).norm() / T.norm() < tol, f"Failed to converge for non-sequential option"

    impossible_tol = 1e-20
    with pytest.warns(UserWarning):
        result = ttb.hosvd(T, impossible_tol)
    assert (
        result.full() - T
    ).norm() / T.norm() > impossible_tol, f"Converged beyond provided precision"


@pytest.mark.indevelopment
def test_hosvd_default_init(capsys, sample_tensor):
    (data, T) = sample_tensor
    _ = ttb.hosvd(T, 1)


@pytest.mark.indevelopment
def test_hosvd_smoke_test_verbosity(capsys, sample_tensor):
    """For now just make sure verbosity calcs don't crash"""
    (data, T) = sample_tensor
    ttb.hosvd(T, 1, verbosity=10)


@pytest.mark.indevelopment
def test_hosvd_incorrect_ranks(capsys, sample_tensor):
    (data, T) = sample_tensor
    ranks = list(range(T.ndims - 1))
    with pytest.raises(ValueError):
        _ = ttb.hosvd(T, 1, ranks=ranks)


@pytest.mark.indevelopment
def test_hosvd_incorrect_dimorder(capsys, sample_tensor):
    (data, T) = sample_tensor
    dimorder = list(range(T.ndims - 1))
    with pytest.raises(ValueError):
        _ = ttb.hosvd(T, 1, dimorder=dimorder)

    dimorder = 1
    with pytest.raises(ValueError):
        _ = ttb.hosvd(T, 1, dimorder=dimorder)


@pytest.mark.indevelopment
def test_hosvd_3way(capsys, sample_tensor_3way):
    (data, T) = sample_tensor_3way
    M = ttb.hosvd(T, 1e-4, verbosity=0)
    capsys.readouterr()
    print(f"M=\n{M}")
    core = np.array(
        [
            [
                [-8.301598119750199e01, -5.005881796972034e-03],
                [-1.268039597172832e-02, 5.842630378620833e00],
            ],
            [
                [3.709974006281391e-02, -1.915213813096568e00],
                [-5.157111619887230e-01, 5.243776123493664e-01],
            ],
        ]
    )
    fm0 = np.array(
        [
            [-5.452132631706279e-01, -7.321719955012304e-01],
            [-5.767748638548937e-01, -2.576993904719336e-02],
            [-6.083364645391598e-01, 6.806321174064961e-01],
        ]
    )
    fm1 = np.array(
        [
            [-4.756392343758577e-01, 7.791666394653051e-01],
            [-5.719678320081717e-01, 7.865197061237804e-02],
            [-6.682964296404851e-01, -6.218626982406427e-01],
        ]
    )
    fm2 = np.array(
        [
            [-1.922305666539489e-01, 8.924016710972924e-01],
            [-5.140779746206554e-01, 2.627873081852611e-01],
            [-8.359253825873615e-01, -3.668270547267537e-01],
        ]
    )
    expected = ttb.ttensor.from_data(ttb.tensor.from_data(core), [fm0, fm1, fm2])
    assert np.allclose(M.double(), expected.double())
    assert np.allclose(np.abs(M.core.data), np.abs(core))
    assert np.allclose(np.abs(M.u[0]), np.abs(fm0))
    assert np.allclose(np.abs(M.u[1]), np.abs(fm1))
    assert np.allclose(np.abs(M.u[2]), np.abs(fm2))



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import logging

import numpy as np
import pytest
import scipy.sparse as sparse

import pyttb as ttb


@pytest.mark.indevelopment
def test_sptensor_to_dense_matrix():
    subs = np.array([[1, 1, 1], [1, 1, 3], [2, 2, 2], [3, 3, 3]])
    vals = np.array([[0.5], [1.5], [2.5], [3.5]])
    shape = (4, 4, 4)
    mode0 = sparse.coo_matrix(
        ([0.5, 1.5, 2.5, 3.5], ([5, 13, 10, 15], [1, 1, 2, 3]))
    ).toarray()
    mode1 = sparse.coo_matrix(
        ([0.5, 1.5, 2.5, 3.5], ([5, 13, 10, 15], [1, 1, 2, 3]))
    ).toarray()
    mode2 = sparse.coo_matrix(
        ([0.5, 1.5, 2.5, 3.5], ([5, 5, 10, 15], [1, 3, 2, 3]))
    ).toarray()
    Ynt = [mode0, mode1, mode2]

    sptensorInstance = ttb.sptensor().from_data(subs, vals, shape)
    tensorInstance = sptensorInstance.full()

    for mode in range(sptensorInstance.ndims):
        Xnt = ttb.tt_to_dense_matrix(tensorInstance, mode, True)
        assert np.array_equal(Xnt, Ynt[mode])


@pytest.mark.indevelopment
def test_sptensor_from_dense_matrix():
    tensorInstance = ttb.tensor.from_data(np.random.normal(size=(4, 4, 4)))
    for mode in range(tensorInstance.ndims):
        tensorCopy = ttb.tensor.from_tensor_type(tensorInstance)
        Xnt = ttb.tt_to_dense_matrix(tensorCopy, mode, True)
        Ynt = ttb.tt_from_dense_matrix(Xnt, tensorCopy.shape, mode, 0)
        assert tensorCopy.isequal(Ynt)

    for mode in range(tensorInstance.ndims):
        tensorCopy = ttb.tensor.from_tensor_type(tensorInstance)
        Xnt = ttb.tt_to_dense_matrix(tensorCopy, mode, False)
        Ynt = ttb.tt_from_dense_matrix(Xnt, tensorCopy.shape, mode, 1)
        assert tensorCopy.isequal(Ynt)


@pytest.mark.indevelopment
def test_tt_union_rows():
    a = np.array([[4, 6], [1, 9], [2, 6], [2, 6], [99, 0]])
    b = np.array([[1, 7], [1, 8], [2, 6]])
    assert (
        ttb.tt_union_rows(a, b)
        == np.array([[1, 7], [1, 8], [4, 6], [1, 9], [2, 6], [99, 0]])
    ).all()
    _, idx = np.unique(a, axis=0, return_index=True)
    assert (ttb.tt_union_rows(a, np.array([])) == a[np.sort(idx)]).all()
    assert (ttb.tt_union_rows(np.array([]), a) == a[np.sort(idx)]).all()


@pytest.mark.indevelopment
def test_tt_dimscheck():
    #  Empty
    rdims, ridx = ttb.tt_dimscheck(6, dims=np.array([]))
    assert (rdims == np.array([0, 1, 2, 3, 4, 5])).all()
    assert ridx is None

    # Exclude Dims
    rdims, ridx = ttb.tt_dimscheck(6, exclude_dims=np.array([1]))
    assert (rdims == np.array([0, 2, 3, 4, 5])).all()
    assert ridx is None

    # Invalid minus
    with pytest.raises(ValueError) as excinfo:
        ttb.tt_dimscheck(6, 6, exclude_dims=np.array([7]))
    assert "Exclude dims" in str(excinfo)

    # Positive
    rdims, ridx = ttb.tt_dimscheck(6, dims=np.array([5]))
    assert (rdims == np.array([5])).all()
    assert ridx is None

    # M==P
    rdims, ridx = ttb.tt_dimscheck(6, 5, exclude_dims=np.array([0]))
    assert (rdims == np.array([1, 2, 3, 4, 5])).all()
    assert (ridx == np.arange(0, 5)).all()

    # M==N
    rdims, ridx = ttb.tt_dimscheck(6, 6, exclude_dims=np.array([0]))
    assert (rdims == np.array([1, 2, 3, 4, 5])).all()
    assert (ridx == rdims).all()

    # M>N
    with pytest.raises(AssertionError) as excinfo:
        ttb.tt_dimscheck(6, 7, exclude_dims=np.array([0]))
    assert "Cannot have more multiplicands than dimensions" in str(excinfo)

    # M!=P and M!=N
    with pytest.raises(AssertionError) as excinfo:
        ttb.tt_dimscheck(6, 4, exclude_dims=np.array([0]))
    assert "Invalid number of multiplicands" in str(excinfo)

    # Both dims and exclude dims
    with pytest.raises(ValueError) as excinfo:
        ttb.tt_dimscheck(6, dims=[], exclude_dims=[])
    assert "not both" in str(excinfo)

    # We no longer support negative dims. Make sure that is explicit
    with pytest.raises(ValueError) as excinfo:
        ttb.tt_dimscheck(6, dims=np.array([-1]))
    assert "Negative dims" in str(excinfo), f"{str(excinfo)}"


@pytest.mark.indevelopment
def test_tt_tenfun():
    data = np.array([[1, 2, 3], [4, 5, 6]])
    t1 = ttb.tensor.from_data(data)
    t2 = ttb.tensor.from_data(data)

    # Binary case
    def add(x, y):
        return x + y

    assert (ttb.tt_tenfun(add, t1, t2).data == 2 * data).all()

    # Single argument case
    def add1(x):
        return x + 1

    assert (ttb.tt_tenfun(add1, t1).data == (data + 1)).all()

    # Multi argument case
    def tensor_max(x):
        return np.max(x, axis=0)

    assert (ttb.tt_tenfun(tensor_max, t1, t1, t1).data == data).all()
    # TODO: sptensor arguments, depends on fixing the indexing ordering

    # No np array case
    assert (ttb.tt_tenfun(tensor_max, data, data, data).data == data).all()

    # No argument case
    with pytest.raises(AssertionError) as excinfo:
        ttb.tt_tenfun(tensor_max)
    assert "Must provide element(s) to perform operation on" in str(excinfo)

    # No list case
    with pytest.raises(AssertionError) as excinfo:
        ttb.tt_tenfun(tensor_max, [1, 2, 3])
    assert "Invalid input to ten fun" in str(excinfo)

    # Scalar argument not in first two positions
    with pytest.raises(AssertionError) as excinfo:
        ttb.tt_tenfun(tensor_max, t1, t1, 1)
    assert "Argument 2 is a scalar but expected a tensor" in str(excinfo)

    # Tensors of different sizes
    with pytest.raises(AssertionError) as excinfo:
        ttb.tt_tenfun(
            tensor_max,
            t1,
            t1,
            ttb.tensor.from_data(np.concatenate((data, np.array([[7, 8, 9]])))),
        )
    assert "Tensor 2 is not the same size as the first tensor input" in str(excinfo)


@pytest.mark.indevelopment
def test_tt_setdiff_rows():
    a = np.array([[4, 6], [1, 9], [2, 6], [2, 6], [99, 0]])
    b = np.array(
        [[1, 7], [1, 8], [2, 6], [2, 1], [2, 4], [4, 6], [4, 7], [5, 9], [5, 2], [5, 1]]
    )
    assert (ttb.tt_setdiff_rows(a, b) == np.array([1, 4])).all()

    a = np.array([[4, 6], [1, 9]])
    b = np.array([[1, 7], [1, 8]])
    assert (ttb.tt_setdiff_rows(a, b) == np.array([0, 1])).all()

    a = np.array([[4, 6], [1, 9]])
    b = np.array([])
    assert (ttb.tt_setdiff_rows(a, b) == np.arange(a.shape[0])).all()
    assert (ttb.tt_setdiff_rows(b, a) == b).all()


@pytest.mark.indevelopment
def test_tt_intersect_rows():
    a = np.array([[4, 6], [1, 9], [2, 6], [2, 6]])
    b = np.array(
        [[1, 7], [1, 8], [2, 6], [2, 1], [2, 4], [4, 6], [4, 7], [5, 9], [5, 2], [5, 1]]
    )
    assert (ttb.tt_intersect_rows(a, b) == np.array([2, 0])).all()

    a = np.array([[4, 6], [1, 9]])
    b = np.array([])
    assert (ttb.tt_intersect_rows(a, b) == b).all()
    assert (ttb.tt_intersect_rows(a, b) == ttb.tt_intersect_rows(b, a)).all()


@pytest.mark.indevelopment
def test_tt_ismember_rows():
    a = np.array([[4, 6], [1, 9], [2, 6]])
    b = np.array(
        [[1, 7], [1, 8], [2, 6], [2, 1], [2, 4], [4, 6], [4, 7], [5, 9], [5, 2], [5, 1]]
    )
    assert (ttb.tt_ismember_rows(a, b) == np.array([5, -1, 2])).all()
    assert (
        ttb.tt_ismember_rows(b, a) == np.array([-1, -1, 2, -1, -1, 0, -1, -1, -1, -1])
    ).all()


@pytest.mark.indevelopment
def test_tt_irenumber():
    # TODO: Note this is more of a regression test by exploring the behaviour in MATLAB still not totally clear on WHY it behaves this way
    # Constant shouldn't effect performance
    const = 1

    subs = np.array([[const, const, 0], [const, const, 1]])
    vals = np.array([[0.5], [1.5]])
    shape = (4, 4, 4)
    data = {"subs": subs, "vals": vals, "shape": shape}
    sptensorInstance = ttb.sptensor().from_data(subs, vals, shape)
    slice_tuple = (
        slice(None, None, None),
        slice(None, None, None),
        slice(None, None, None),
    )
    extended_result = np.array(
        [[const, const, const, const, const, 0], [const, const, const, const, const, 1]]
    )

    # Pad equal to number of modes
    assert (
        ttb.tt_irenumber(sptensorInstance, shape, (const, const, const))
        == extended_result
    ).all()

    # Full slice should equal original
    assert (ttb.tt_irenumber(sptensorInstance, shape, slice_tuple) == subs).all()

    # Verify that list and equivalent slice act the same
    assert (
        ttb.tt_irenumber(sptensorInstance, shape, (const, const, slice(0, 1, 1)))
        == np.array([[const, const, const, const, 0], [const, const, const, const, 1]])
    ).all()
    assert (
        ttb.tt_irenumber(sptensorInstance, shape, (const, const, [0, 1]))
        == np.array([[const, const, const, const, 0], [const, const, const, const, 1]])
    ).all()


@pytest.mark.indevelopment
def test_tt_assignment_type():
    # type(x)==type(rhs)
    x = 5
    rhs = 5
    subs = 5
    assert ttb.tt_assignment_type(x, subs, rhs) == "subtensor"

    # type(x)!=type(rhs), subs dimensionality >=2
    rhs = "cat"
    subs = (1, 1, 1)
    assert ttb.tt_assignment_type(x, subs, rhs) == "subtensor"

    subs = (np.array([1, 2, 3]),)
    assert ttb.tt_assignment_type(x, subs, rhs) == "subscripts"

    # type(x)!=type(rhs), subs dimensionality <2
    subs = np.array([1])
    assert ttb.tt_assignment_type(x, subs, rhs) == "subscripts"


@pytest.mark.indevelopment
def test_tt_renumberdims():
    # Singleton output
    shape = 5
    number_range = 1
    idx = 3
    assert ttb.tt_renumberdim(idx, shape, number_range) == (0, 0)

    # Array output
    shape = 5
    number_range = np.array([1, 3, 4])
    idx = [1, 3, 4]
    newidx, newshape = ttb.tt_renumberdim(idx, shape, number_range)
    assert (newidx == np.array([0, 1, 2])).all()
    assert newshape == 3

    # Slice input
    shape = 5
    number_range = slice(1, 3, None)
    idx = [1, 2]
    newidx, newshape = ttb.tt_renumberdim(idx, shape, number_range)
    assert (newidx == np.array([0, 1])).all()
    assert newshape == 2


@pytest.mark.indevelopment
def test_tt_renumber():
    # Singleton in each dimension
    shape = (5, 5, 5)
    number_range = (1, 1, 1)
    subs = np.array([[3, 3, 3]])
    newsubs, newshape = ttb.tt_renumber(subs, shape, number_range)
    assert (newsubs == np.array([[0, 0, 0]])).all()
    assert newshape == (0, 0, 0)

    # Array in each dimension
    shape = (5, 5, 5)
    number_range = ([1, 3, 4], [1, 3, 4], [1, 3, 4])
    subs = np.array([[1, 1, 1], [3, 3, 3]])
    newsubs, newshape = ttb.tt_renumber(subs, shape, number_range)
    assert (newsubs == np.array([[0, 0, 0], [1, 1, 1]])).all()
    assert newshape == (3, 3, 3)

    # Slice in each dimension
    shape = (5, 5, 5)
    number_range = (slice(1, 3, None), slice(1, 3, None), slice(1, 3, None))
    subs = np.array([[2, 2, 2]])
    newsubs, newshape = ttb.tt_renumber(subs, shape, number_range)
    assert (newsubs == np.array([[1, 1, 1]])).all()
    assert newshape == (2, 2, 2)

    # Slice in each dimension, empty subs
    shape = (5, 5, 5)
    number_range = (slice(1, 3, None), slice(1, 3, None), slice(1, 3, None))
    subs = np.array([])
    newsubs, newshape = ttb.tt_renumber(subs, shape, number_range)
    assert newsubs.size == 0
    assert newshape == (2, 2, 2)

    # Not slice in each dimension, empty subs
    shape = (5, 5, 5)
    number_range = ([1, 3, 4], [1, 3, 4], [1, 3, 4])
    subs = np.array([])
    newsubs, newshape = ttb.tt_renumber(subs, shape, number_range)
    assert newsubs.size == 0
    assert newshape == (3, 3, 3)


@pytest.mark.indevelopment
def test_tt_sub2ind_valid():
    subs = np.array([[0, 0, 0], [1, 1, 1], [3, 3, 3]])
    idx = np.array([0, 21, 63])
    siz = np.array([4, 4, 4])
    assert (ttb.tt_sub2ind(siz, subs) == idx).all()

    empty = np.array([])
    assert (ttb.tt_sub2ind(siz, empty) == empty).all()


@pytest.mark.indevelopment
def test_tt_ind2sub_valid():
    subs = np.array([[0, 0, 0], [1, 1, 1], [3, 3, 3]])
    idx = np.array([0, 21, 63])
    shape = (4, 4, 4)
    logging.debug(f"\nttb.tt_ind2sub(shape, idx): {ttb.tt_ind2sub(shape, idx)}")
    assert (ttb.tt_ind2sub(shape, idx) == subs).all()

    subs = np.array([[1, 0], [0, 1]])
    idx = np.array([1, 2])
    shape = (2, 2)
    logging.debug(f"\nttb.tt_ind2sub(shape, idx): {ttb.tt_ind2sub(shape, idx)}")
    assert (ttb.tt_ind2sub(shape, idx) == subs).all()

    empty = np.array([])
    assert (
        ttb.tt_ind2sub(shape, empty) == np.empty(shape=(0, len(shape)), dtype=int)
    ).all()


@pytest.mark.indevelopment
def test_tt_subsubsref_valid():
    subs = np.array([[0, 0, 0], [1, 1, 1], [3, 3, 3]])
    vals = np.array([[0], [21], [63]])
    shape = (4, 4, 4)
    a = ttb.sptensor.from_data(subs, vals, shape)
    assert isinstance(ttb.tt_subsubsref(a, [1]), ttb.sptensor)
    # TODO need to understand behavior better
    assert True


@pytest.mark.indevelopment
def test_tt_intvec2str_valid():
    """This function is slotted to be removed because it is probably unnecessary in python"""
    v = np.array([1, 2, 3])
    assert ttb.tt_intvec2str(v) == "[1 2 3]"


@pytest.mark.indevelopment
def test_tt_sizecheck_empty():
    assert ttb.tt_sizecheck(())


@pytest.mark.indevelopment
def test_tt_sizecheck_valid():
    assert ttb.tt_sizecheck((2, 2, 2))


@pytest.mark.indevelopment
def test_tt_sizecheck_invalid():
    # Float
    assert not ttb.tt_sizecheck((1.0, 2, 2))
    # Too many dimensions
    assert not ttb.tt_sizecheck(np.array([[2, 2], [2, 2]]))
    # Nan
    assert not ttb.tt_sizecheck((np.nan, 2, 2))
    # Inf
    assert not ttb.tt_sizecheck((np.inf, 2, 2))
    # Zero
    assert not ttb.tt_sizecheck((0, 2, 2))


@pytest.mark.indevelopment
def test_tt_sizecheck_errorMessage():
    # Raise when nargout == 0
    with pytest.raises(AssertionError) as excinfo:
        ttb.tt_sizecheck((1.0, 2, 2), nargout=False)
    assert "Size must be a row vector of real positive integers" in str(excinfo)


@pytest.mark.indevelopment
def test_tt_subscheck_empty():
    assert ttb.tt_subscheck(np.array([]))


@pytest.mark.indevelopment
def test_tt_subscheck_valid():
    assert ttb.tt_subscheck(np.array([[2, 2], [2, 2]]))


@pytest.mark.indevelopment
def test_tt_subscheck_invalid():
    # Too Few Dims
    assert not ttb.tt_subscheck(np.array([2]))
    # Too Many Dims
    assert not ttb.tt_subscheck(np.array([[[2]]]))
    # Less than 0
    assert not ttb.tt_subscheck(np.array([[-1, 2], [2, 2]]))
    # Nan
    assert not ttb.tt_subscheck(np.array([[np.nan, 2], [2, 2]]))
    # Inf
    assert not ttb.tt_subscheck(np.array([[np.inf, 2], [2, 2]]))
    # Non-int
    assert not ttb.tt_subscheck(np.array([[1.0, 2], [2, 2]]))


@pytest.mark.indevelopment
def test_tt_subscheck_errorMessage():
    # Raise when nargout == 0
    with pytest.raises(AssertionError) as excinfo:
        ttb.tt_subscheck(np.array([1.0, 2, 2]), nargout=False)
    assert "Subscripts must be a matrix of real positive integers" in str(excinfo)


@pytest.mark.indevelopment
def test_tt_valscheck_empty():
    assert ttb.tt_valscheck(np.array([]))


@pytest.mark.indevelopment
def test_tt_valscheck_valid():
    assert ttb.tt_valscheck(np.array([[0.5], [1.5], [2.5]]))


@pytest.mark.indevelopment
def test_tt_valscheck_invalid():
    # Row array
    assert not ttb.tt_valscheck(np.array([2, 2]))
    # Matrix, too many dimensions
    assert not ttb.tt_valscheck(np.array([[2, 2], [2, 2]]))


@pytest.mark.indevelopment
def test_tt_valscheck_errorMessage():
    # Raise when nargout == 0
    with pytest.raises(AssertionError) as excinfo:
        ttb.tt_valscheck(np.array([2, 2]), nargout=False)
    assert "Values must be in array" in str(excinfo)


@pytest.mark.indevelopment
def test_isrow_empty():
    assert not ttb.isrow(np.array([[]]))


@pytest.mark.indevelopment
def test_isrow_valid():
    assert ttb.isrow(np.array([[2, 2, 2]]))


@pytest.mark.indevelopment
def test_isrow_invalid():
    # 2 x 2 Matrix
    assert not ttb.isrow(np.array([[2, 2], [2, 2]]))
    # Column vector
    assert not ttb.isrow(np.array([[2, 2, 2]]).T)


@pytest.mark.indevelopment
def test_isvector_empty():
    assert ttb.isvector(np.array([[]]))


@pytest.mark.indevelopment
def test_isvector_valid():
    assert ttb.isvector(np.array([[2, 2, 2]]))
    assert ttb.isvector(np.array([[2, 2, 2]]).T)


@pytest.mark.indevelopment
def test_isvector_invalid():
    # 2 x 2 Matrix
    assert not ttb.isvector(np.array([[2, 2], [2, 2]]))


@pytest.mark.indevelopment
def test_islogical_empty():
    assert not ttb.islogical(np.array([[]]))


@pytest.mark.indevelopment
def test_islogical_valid():
    assert ttb.islogical(True)


@pytest.mark.indevelopment
def test_islogical_invalid():
    assert not ttb.islogical(np.array([[2, 2, 2]]))
    assert not ttb.islogical(1.1)
    assert not ttb.islogical(0)



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import numpy as np
import pytest
import scipy.sparse as sparse

import pyttb as ttb


@pytest.fixture()
def sample_ttensor():
    """Simple TTENSOR to verify by hand"""
    core = ttb.tensor.from_data(np.ones((2, 2, 2)))
    factors = [np.ones((1, 2))] * len(core.shape)
    ttensorInstance = ttb.ttensor().from_data(core, factors)
    return ttensorInstance


@pytest.fixture()
def random_ttensor():
    """Arbitrary TTENSOR to verify consistency between alternative operations"""
    core = ttb.tensor.from_data(np.random.random((2, 3, 4)))
    factors = [
        np.random.random((5, 2)),
        np.random.random((2, 3)),
        np.random.random((4, 4)),
    ]
    ttensorInstance = ttb.ttensor().from_data(core, factors)
    return ttensorInstance


@pytest.mark.indevelopment
def test_ttensor_initialization_empty():
    empty_tensor = ttb.tensor()

    # No args
    ttensorInstance = ttb.ttensor()
    assert ttensorInstance.core == empty_tensor
    assert ttensorInstance.u == []


@pytest.mark.indevelopment
def test_ttensor_initialization_from_data(sample_ttensor):
    ttensorInstance = sample_ttensor
    assert isinstance(ttensorInstance.core, ttb.tensor)
    assert all([isinstance(a_factor, np.ndarray) for a_factor in ttensorInstance.u])

    # Negative Tests
    non_array_factor = ttensorInstance.u + [1]
    with pytest.raises(ValueError):
        ttb.ttensor.from_data(ttensorInstance.core, non_array_factor[1:])

    non_matrix_factor = ttensorInstance.u + [np.array([1])]
    with pytest.raises(ValueError):
        ttb.ttensor.from_data(ttensorInstance.core, non_matrix_factor[1:])

    too_few_factors = ttensorInstance.u.copy()
    too_few_factors.pop()
    with pytest.raises(ValueError):
        ttb.ttensor.from_data(ttensorInstance.core, too_few_factors)

    wrong_shape_factor = ttensorInstance.u.copy()
    row, col = wrong_shape_factor[0].shape
    wrong_shape_factor[0] = np.random.random((row + 1, col + 1))
    with pytest.raises(ValueError):
        ttb.ttensor.from_data(ttensorInstance.core, wrong_shape_factor)

    # Enforce error until sptensor core/other cores supported
    with pytest.raises(ValueError):
        ttb.ttensor.from_data(
            ttb.sptensor.from_tensor_type(ttensorInstance.core), ttensorInstance.u
        )


@pytest.mark.indevelopment
def test_ttensor_initialization_from_tensor_type(sample_ttensor):
    # Copy constructor
    ttensorInstance = sample_ttensor
    ttensorCopy = ttb.ttensor.from_tensor_type(ttensorInstance)
    assert ttensorCopy.core == ttensorInstance.core
    assert ttensorCopy.u == ttensorInstance.u
    assert ttensorCopy.shape == ttensorInstance.shape


@pytest.mark.indevelopment
def test_ttensor_full(sample_ttensor):
    ttensorInstance = sample_ttensor
    tensor = ttensorInstance.full()
    # This sanity check only works for all 1's
    assert tensor.double() == np.prod(ttensorInstance.core.shape)

    # Negative tests
    sparse_core = ttb.sptensor()
    sparse_core.shape = ttensorInstance.core.shape
    sparse_u = [
        sparse.coo_matrix(np.zeros(factor.shape)) for factor in ttensorInstance.u
    ]
    # We could probably make these properties to avoid this edge case but expect to eventually cover these alternate
    # cores
    ttensorInstance.core = sparse_core
    ttensorInstance.u = sparse_u
    with pytest.raises(ValueError):
        ttensorInstance.full()


@pytest.mark.indevelopment
def test_ttensor_double(sample_ttensor):
    ttensorInstance = sample_ttensor
    # This sanity check only works for all 1's
    assert ttensorInstance.double() == np.prod(ttensorInstance.core.shape)


@pytest.mark.indevelopment
def test_ttensor_ndims(sample_ttensor):
    ttensorInstance = sample_ttensor

    assert ttensorInstance.ndims == 3


@pytest.mark.indevelopment
def test_ttensor__pos__(sample_ttensor):
    ttensorInstance = sample_ttensor
    ttensorInstance2 = +ttensorInstance

    assert ttensorInstance.isequal(ttensorInstance2)


@pytest.mark.indevelopment
def test_sptensor__neg__(sample_ttensor):
    ttensorInstance = sample_ttensor
    ttensorInstance2 = -ttensorInstance
    ttensorInstance3 = -ttensorInstance2

    assert not ttensorInstance.isequal(ttensorInstance2)
    assert ttensorInstance.isequal(ttensorInstance3)


@pytest.mark.indevelopment
def test_ttensor_innerproduct(sample_ttensor, random_ttensor):
    ttensorInstance = sample_ttensor

    # TODO these are an overly simplistic edge case for ttensors that are a single float

    # ttensor innerprod ttensor
    assert ttensorInstance.innerprod(ttensorInstance) == ttensorInstance.double() ** 2
    core_dim = ttensorInstance.core.shape[0] + 1
    ndim = ttensorInstance.ndims
    large_core_ttensor = ttb.ttensor.from_data(
        ttb.tensor.from_data(np.ones((core_dim,) * ndim)),
        [np.ones((1, core_dim))] * ndim,
    )
    assert large_core_ttensor.innerprod(
        ttensorInstance
    ) == ttensorInstance.full().innerprod(large_core_ttensor.full())

    # ttensor innerprod tensor
    assert (
        ttensorInstance.innerprod(ttensorInstance.full())
        == ttensorInstance.double() ** 2
    )

    # ttensr innerprod ktensor
    ktensorInstance = ttb.ktensor.from_data(np.array([8.0]), [np.array([[1.0]])] * 3)
    assert ttensorInstance.innerprod(ktensorInstance) == ttensorInstance.double() ** 2

    # ttensor innerprod tensor (shape larger than core)
    random_ttensor.innerprod(random_ttensor.full())

    # Negative Tests
    ttensor_extra_factors = ttb.ttensor.from_tensor_type(ttensorInstance)
    ttensor_extra_factors.u.extend(ttensorInstance.u)
    with pytest.raises(ValueError):
        ttensorInstance.innerprod(ttensor_extra_factors)

    tensor_extra_dim = ttb.tensor.from_data(np.ones(ttensorInstance.shape + (1,)))
    with pytest.raises(ValueError):
        ttensorInstance.innerprod(tensor_extra_dim)

    invalid_option = []
    with pytest.raises(ValueError):
        ttensorInstance.innerprod(invalid_option)


@pytest.mark.indevelopment
def test_ttensor__mul__(sample_ttensor):
    ttensorInstance = sample_ttensor
    mul_factor = 2

    # This sanity check only works for all 1's
    assert (ttensorInstance * mul_factor).double() == np.prod(
        ttensorInstance.core.shape
    ) * mul_factor
    assert (ttensorInstance * float(2)).double() == np.prod(
        ttensorInstance.core.shape
    ) * float(mul_factor)

    # Negative tests
    with pytest.raises(ValueError):
        _ = ttensorInstance * "some_string"


@pytest.mark.indevelopment
def test_ttensor__rmul__(sample_ttensor):
    ttensorInstance = sample_ttensor
    mul_factor = 2

    # This sanity check only works for all 1's
    assert (mul_factor * ttensorInstance).double() == np.prod(
        ttensorInstance.core.shape
    ) * mul_factor
    assert (float(2) * ttensorInstance).double() == np.prod(
        ttensorInstance.core.shape
    ) * float(mul_factor)

    # Negative tests
    with pytest.raises(ValueError):
        _ = "some_string" * ttensorInstance


@pytest.mark.indevelopment
def test_ttensor_ttv(sample_ttensor):
    ttensorInstance = sample_ttensor
    mul_factor = 1
    trivial_vectors = [np.array([mul_factor])] * len(ttensorInstance.shape)
    final_value = sample_ttensor.ttv(trivial_vectors)
    assert final_value == np.prod(ttensorInstance.core.shape)

    assert np.allclose(
        ttensorInstance.ttv(trivial_vectors[0], 0).double(),
        ttensorInstance.full().ttv(trivial_vectors[0], 0).double(),
    )

    assert np.allclose(
        ttensorInstance.ttv(trivial_vectors[0:2], exclude_dims=0).double(),
        ttensorInstance.full().ttv(trivial_vectors[0:2], exclude_dims=0).double(),
    )

    # Negative tests
    wrong_shape_vector = trivial_vectors.copy()
    wrong_shape_vector[0] = np.array([mul_factor, mul_factor])
    with pytest.raises(ValueError):
        sample_ttensor.ttv(wrong_shape_vector)


@pytest.mark.indevelopment
def test_ttensor_mttkrp(random_ttensor):
    ttensorInstance = random_ttensor
    column_length = 6
    vectors = [np.random.random((u.shape[0], column_length)) for u in ttensorInstance.u]
    final_value = ttensorInstance.mttkrp(vectors, 2)
    full_value = ttensorInstance.full().mttkrp(vectors, 2)
    assert np.allclose(final_value, full_value), (
        f"TTensor value is: \n{final_value}\n\n" f"Full value is: \n{full_value}"
    )


@pytest.mark.indevelopment
def test_ttensor_norm(sample_ttensor, random_ttensor):
    ttensorInstance = random_ttensor
    assert np.isclose(ttensorInstance.norm(), ttensorInstance.full().norm())

    # Core larger than full tensor
    ttensorInstance = sample_ttensor
    assert np.isclose(ttensorInstance.norm(), ttensorInstance.full().norm())


@pytest.mark.indevelopment
def test_ttensor_permute(random_ttensor):
    ttensorInstance = random_ttensor
    original_order = np.arange(0, len(ttensorInstance.core.shape))
    permuted_tensor = ttensorInstance.permute(original_order)
    assert ttensorInstance.isequal(permuted_tensor)

    # Negative Tests
    with pytest.raises(ValueError):
        bad_permutation_order = np.arange(0, len(ttensorInstance.core.shape) + 1)
        ttensorInstance.permute(bad_permutation_order)


@pytest.mark.indevelopment
def test_ttensor_ttm(random_ttensor):
    ttensorInstance = random_ttensor
    row_length = 9
    matrices = [np.random.random((row_length, u.shape[0])) for u in ttensorInstance.u]
    final_value = ttensorInstance.ttm(matrices, np.arange(len(matrices)))
    reverse_value = ttensorInstance.ttm(
        list(reversed(matrices)), np.arange(len(matrices) - 1, -1, -1)
    )
    assert final_value.isequal(reverse_value), (
        f"TTensor value is: \n{final_value}\n\n" f"Full value is: \n{reverse_value}"
    )
    final_value = ttensorInstance.ttm(matrices)  # No dims
    assert final_value.isequal(reverse_value)
    final_value = ttensorInstance.ttm(
        matrices, list(range(len(matrices)))
    )  # Dims as list
    assert final_value.isequal(reverse_value)

    # Exclude Dims
    assert ttensorInstance.ttm(matrices[1:], exclude_dims=0).isequal(
        ttensorInstance.ttm(matrices[1:], dims=np.array([1, 2]))
    )

    single_tensor_result = ttensorInstance.ttm(matrices[0], 0)
    single_tensor_full_result = ttensorInstance.full().ttm(matrices[0], 0)
    assert np.allclose(
        single_tensor_result.double(), single_tensor_full_result.double()
    ), (
        f"TTensor value is: \n{single_tensor_result.full()}\n\n"
        f"Full value is: \n{single_tensor_full_result}"
    )

    transposed_matrices = [matrix.transpose() for matrix in matrices]
    transpose_value = ttensorInstance.ttm(
        transposed_matrices, np.arange(len(matrices)), transpose=True
    )
    assert final_value.isequal(transpose_value)

    # Negative Tests
    big_wrong_size = 123
    bad_matrices = matrices.copy()
    bad_matrices[0] = np.random.random((big_wrong_size, big_wrong_size))
    with pytest.raises(ValueError):
        _ = ttensorInstance.ttm(bad_matrices, np.arange(len(bad_matrices)))

    with pytest.raises(ValueError):
        # Negative dims currently broken, ensure we catch early and
        # remove once resolved
        ttensorInstance.ttm(matrices, -1)


@pytest.mark.indevelopment
def test_ttensor_reconstruct(random_ttensor):
    ttensorInstance = random_ttensor
    # TODO: This slice drops the singleton dimension, should it? If so should ttensor squeeze during reconstruct?
    full_slice = ttensorInstance.full()[:, 1, :]
    ttensor_slice = ttensorInstance.reconstruct(1, 1)
    assert np.allclose(full_slice.double(), ttensor_slice.squeeze().double())
    assert ttensorInstance.reconstruct().isequal(ttensorInstance.full())
    sample_all_modes = [np.array([0])] * len(ttensorInstance.shape)
    sample_all_modes[-1] = 0  # Make raw scalar
    reconstruct_scalar = ttensorInstance.reconstruct(sample_all_modes).full().double()
    full_scalar = ttensorInstance.full()[tuple(sample_all_modes)]
    assert np.isclose(reconstruct_scalar, full_scalar)

    scale = np.random.random(ttensorInstance.u[1].shape).transpose()
    _ = ttensorInstance.reconstruct(scale, 1)
    # FIXME from the MATLAB docs wasn't totally clear how to validate this

    # Negative Tests
    with pytest.raises(ValueError):
        _ = ttensorInstance.reconstruct(1, [0, 1])


@pytest.mark.indevelopment
def test_ttensor_nvecs(random_ttensor):
    ttensorInstance = random_ttensor
    n = 0
    r = 2
    ttensor_eigvals = ttensorInstance.nvecs(n, r)
    full_eigvals = ttensorInstance.full().nvecs(n, r)
    assert np.allclose(ttensor_eigvals, full_eigvals)

    # Test for eig vals larger than shape-1
    n = 1
    r = 2
    full_eigvals = ttensorInstance.full().nvecs(n, r)
    with pytest.warns(Warning) as record:
        ttensor_eigvals = ttensorInstance.nvecs(n, r)
    assert (
        "Greater than or equal to tensor.shape[n] - 1 eigenvectors requires cast to dense to solve"
        in str(record[0].message)
    )
    assert np.allclose(ttensor_eigvals, full_eigvals)

    # Negative Tests
    sparse_core = ttb.sptensor()
    sparse_core.shape = ttensorInstance.core.shape
    ttensorInstance.core = sparse_core

    # Sparse core
    with pytest.raises(NotImplementedError):
        ttensorInstance.nvecs(0, 1)

    # Sparse factors
    sparse_u = [
        sparse.coo_matrix(np.zeros(factor.shape)) for factor in ttensorInstance.u
    ]
    ttensorInstance.u = sparse_u
    with pytest.raises(NotImplementedError):
        ttensorInstance.nvecs(0, 1)


@pytest.mark.indevelopment
def test_sptensor_isequal(sample_ttensor):
    ttensorInstance = sample_ttensor
    # Negative Tests
    assert not ttensorInstance.isequal(ttensorInstance.full())
    ttensor_extra_factors = ttb.ttensor.from_tensor_type(ttensorInstance)
    ttensor_extra_factors.u.extend(ttensorInstance.u)
    assert not ttensorInstance.isequal(ttensor_extra_factors)
